import abc
from collections.abc import Iterable, Sequence
from collections import defaultdict
from functools import cache
from typing import Literal

try:
    from typing import override  # py>=3.12
except ImportError:  # pragma: no cover
    from typing_extensions import override

import networkx as nx

from localuf import constants, Repetition
from localuf.decoders.snowflake.constants import RESET, Stage
from localuf.noise import CodeCapacity
from localuf.type_aliases import Edge, Node, Coord
from localuf.constants import Growth
from localuf.decoders.policies import DecodeDrawer
from localuf._base_classes import Code
from localuf._schemes import Frugal
from localuf.decoders._base_uf import direction, BaseUF


class Snowflake(BaseUF):
    """Snowflake decoder based on UF.
    
    Extends ``BaseUF``.
    Incompatible with code capacity noise model.
    Compatible only with frugal scheme.
    
    Does not use:
    
    * ``_growth`` attribute.
    * ``erasure`` attribute.
    """

    BW_DEFAULT_NODE_SIZE = 360
    """The default node size for black-and-white drawing."""

    def __init__(
            self,
            code: Code,
            merger: Literal['fast', 'slow'] = 'fast',
            schedule: Literal['2:1', '1:1'] = '2:1',
            unrooter: Literal['full', 'simple'] = 'full',
            _neighbor_order: Iterable[direction] | None = None,
            _include_timelike_lowest_edges: bool = True,
    ):
        """
        :param code: the code to be decoded.
        :param merger: decides whether nodes flood before syncing (fast) or vice versa (slow) in a merging step.
            Setting this to ``'slow'`` helps break down merging for visualisation.
        :param schedule: the cluster growth schedule.
        :param unrooter: the type of unrooting process to use.
            If ``'full'``, each node in the amputated cluster resets its CID and pointer
            so that the pointer tree structure can be rebuilt from scratch,
            via further merging timesteps.
            If ``'simple'``, the node at breaking point only
            establishes the shortest path to a boundary.
        :param _neighbor_order: optionally customizes the order in which each node checks its neighbors.
            For the repetition code, this should be an iterable of the four directions
            ('W', 'E', 'D', 'U') in some order.
            For the surface code, this should be an iterable of the twelve directions
            ('NWD', 'N', 'NU', 'WD', 'W', 'D', 'U', 'E', 'EU', 'SD', 'S', 'SEU') in some order.
            Technically you can exclude the ones which are not part of the decoding the graph
            e.g. the diagonal directions (comprising >=2 characters) for phenomenological noise,
            but if you accidentally exclude a relevant direction
            then no node will ever check that direction for growing, flooding, nor syncing.
            So it is safest to include all twelve directions even if some are not used.
            If ``_neighbor_order`` is not specified, the orders listed above are used.
        :param _include_timelike_lowest_edges: whether to include
            the purely timelike (i.e. 'up') edges in ``self._LOWEST_EDGES``.
        """
        if isinstance(code.NOISE, CodeCapacity):
            raise ValueError('Snowflake incompatible with code capacity noise model.')
        if not isinstance(code.SCHEME, Frugal):
            raise ValueError('Snowflake only compatible with frugal scheme i.e. `code.SCHEME` must be `Frugal`.')
        super().__init__(code)
        self._SCHEDULE: _Schedule = (_TwoOne if schedule == '2:1' else _OneOne)(self)
        window_height = self.CODE.SCHEME.WINDOW_HEIGHT
        self._EDGES = {
            e: _Edge(self, e) for e in self.CODE.EDGES
            if all(v[self.CODE.TIME_AXIS] < window_height for v in e)
        }
        if _neighbor_order is None:
            if isinstance(code, Repetition):
                _neighbor_order = ('W', 'E', 'D', 'U')
            else:  # Surface
                _neighbor_order = ('NWD', 'N', 'NU', 'WD', 'W', 'D', 'U', 'E', 'EU', 'SD', 'S', 'SEU')
        self._NEIGHBOR_ORDER: Iterable[direction] = _neighbor_order
        self._NODES = {
            v: _Node(
                self,
                v,
                merger=merger,
                unrooter=unrooter,
            ) for v in self.CODE.NODES
            if v[self.CODE.TIME_AXIS] < window_height
        }
        self._DECODE_DRAWER = DecodeDrawer(self._FIG_WIDTH, fig_height=self._FIG_HEIGHT)
        self._stage = Stage.DROP
        """The current stage of the decoder.
        Only used for ``draw_growth`` when ``show_global = True``.
        """
        self._BITSTRING_CONVERTER = (_Repetition if type(code) is Repetition else _Surface)(code.D, window_height)
        self._LOWEST_EDGES = tuple(self.EDGES[e] for e in self._BITSTRING_CONVERTER.lowest_edges(
            str(code.NOISE),
            _include_timelike_lowest_edges=_include_timelike_lowest_edges,
        ))
        """A tuple of the edges in the bottom layer of the viewing window.
        For repetition code, the edges are ordered
        first by type (up, east)
        then by y-coordinate (west to east).
        For surface code, the edges are ordered
        first by type (up, south_down, east_up, south_east_up, south, east),
        then by x-coordinate (north to south), then by y-coordinate (west to east).
        If ``__init__`` was called with ``_include_timelike_lowest_edges = False``,
        the purely timelike (i.e. 'up') edges are excluded.
        """
        self.floor_history: list[str]
        """A list of bitstrings representing
        the correction output from the bottom layer at each drop.
        The order of the bits is given by ``self._LOWEST_EDGES``.
        """
        self.confidence_score_history: defaultdict[str, list[float]] = defaultdict(list)
        """A map from DCS to a list of DCS values after each growth stage."""
    
    def __repr__(self) -> str:
        return f'decoders.Snowflake({self.CODE})'

    @property
    def NODES(self):
        """A map from node coordinates to node objects."""
        return self._NODES

    @property
    def EDGES(self):
        """A map from coordinate pairs to edge objects."""
        return self._EDGES

    @property
    def syndrome(self):
        return self.verbose_syndrome.difference(self.CODE.BOUNDARY_NODES)

    @property
    def verbose_syndrome(self):
        return {v for v, node in self.NODES.items() if node.defect}
    
    @property
    def growth(self):
        return {e: edge.growth for e, edge in self.EDGES.items()}

    @property
    def correction(self):
        return {e for e, edge in self.EDGES.items() if edge.correction}

    @cache
    def index_to_id(self, index: Node):
        """Return ID of node at index (j, t) or (i, j, t)."""
        d = self.CODE.D
        h = self.CODE.SCHEME.WINDOW_HEIGHT
        if len(index) == 2:
            j, t = index
            if j == -1:  # on west boundary
                return 2 * (h-1 - t)
            elif j == d-1:  # on east boundary
                return 2 * (h-1 - t) + 1
            else:
                return 2*h + (d-1)*(h-1 - t) + j
        else:  # 3
            i, j, t = index
            if j == -1:
                return 2*d * (h-1 - t) + i
            elif j == d-1:
                return 2*d * (h-1 - t) + d + i
            else:
                return 2*d*h + d*(d-1)*(h-1 - t) + d*j + i
            
    def init_history(self):
        """Initialize ``history`` and ``floor_history`` attributes."""
        super().init_history()
        self.floor_history: list[str] = []
        
    @cache
    def id_below(self, id_: int):
        """Return ID of node below node with ID ``id_``.
        
        Answer valid only for nodes not on bottom layer of viewing window.
        In implementation this function would be stored in each node.
        """
        d = self.CODE.D
        h = self.CODE.SCHEME.WINDOW_HEIGHT
        if isinstance(self.CODE, Repetition):
            delta = 2 if id_<2*h else d-1
        else:  # Surface
            delta = 2*d if id_<2*d*h else d*(d-1)
        return id_ + delta
        
    @override
    def reset(self):
        super().reset()
        for node in self.NODES.values():
            node.reset()
        for edge in self.EDGES.values():
            edge.reset()
        self._stage = Stage.DROP
        try: del self.history
        except AttributeError: pass
        try: del self.floor_history
        except AttributeError: pass
        self.confidence_score_history = defaultdict(list)
    
    @override
    def decode(
            self,
            syndrome: set[Node],
            log_history: Literal[False, 'fine', 'coarse'] = False,
            log_floor_history: bool = False,
            confidence_scores: Iterable[str] = (),
            noise_level_for_priors: None | float = None,
            time_only: Literal['all', 'merging', 'unrooting'] = 'merging',
            defects_possible: bool = True,
        ):
        """Perform a decoding cycle i.e. a growth round.
        
        
        :param syndrome: The syndrome in the new region discovered by the window raise
            i.e. all defects in ``syndrome`` have the time coordinate ``self.CODE.SCHEME.WINDOW_HEIGHT-1``.
        :param log_history: Whether to populate ``history`` attribute;
            'fine' logs each timestep;
            'coarse', only the final timestep of the growth round.
        :param log_floor_history: Whether to populate ``floor_history`` attribute.
        :param confidence_scores: An iterable of DCS names to record after the decoding cycle.
        :param noise_level_for_priors: Noise level to use when computing some DCSs.
        :param time_only: Whether runtime includes a timestep
            for each drop, each grow, and each merging step ('all');
            each merging step only ('merging');
            or each unrooting step only ('unrooting').
        :param defects_possible: Whether to expect there may be defects in the viewing window
            in the current or any future timestep.
            If ``False``, the decoder will perform only 'drop',
            and will skip 'grow' and 'merge' stages.
            This is useful at the end of the memory experiment after the final syndrome data has come in.
        
        
        :return t: Number of timesteps to complete decoding cycle.
            Equals the increase in ``len(self.history)`` if ``log_history == 'fine'`` and ``time_only == 'all'``.
        """
        self._stage = Stage.DROP
        if log_floor_history:
            self.floor_history.append(
                ''.join(str(int(e.correction)) for e in self._LOWEST_EDGES))
        self.drop(syndrome)
        if log_history == 'fine': self.append_history()
        if defects_possible:
            t = self._SCHEDULE.finish_decode(
                log_history=log_history,
                time_only=time_only,
            )
        else:
            t = 1 if time_only == 'all' else 0
        for confidence_score in confidence_scores:
            if confidence_score == 'runtime':
                self.confidence_score_history['runtime'].append(t)
            elif confidence_score == 'swim_distance':
                swim_distance = self.swim_distance(noise_level=noise_level_for_priors)
                self.confidence_score_history['swim_distance'].append(swim_distance)
            elif confidence_score == 'unclustered_edge_fraction':
                fraction = self.unclustered_edge_fraction(noise_level=noise_level_for_priors)
                self.confidence_score_history['unclustered_edge_fraction'].append(fraction)
            else:
                raise ValueError(f'Unknown confidence score: {confidence_score}')
        return t
    
    def drop(self, syndrome: set[Node]):
        """Make all nodes perform a ``drop`` i.e. raise window by a layer."""
        # SLIDE EDGES DOWN
        for edge in self.EDGES.values():
            edge.CONTACT.drop()
        for edge in self.EDGES.values():
            edge.update_after_drop()
        # SLIDE NODES DOWN
        for node in self.NODES.values():
            node.FRIENDSHIP.drop()
        self._load(syndrome)
        for node in self.NODES.values():
            node.update_after_drop()

    def _load(self, syndrome: set[Node]):
        """Load ``syndrome`` onto decoder.
        
        I.e. top sheet inherits the new measurement round results.
        Note do not set ``active`` as ``defect``
        else nodes in top sheet of viewing window will grow.
        This growth would be premature as there are no edges above these nodes,
        leading to uneven cluster shapes.
        """
        for v in syndrome:
            self.NODES[v].next_defect ^= True

    def merge(
        self,
        whole: bool,
        log_history: Literal[False, 'fine', 'coarse'],
        time_only: Literal['all', 'merging', 'unrooting'] = 'merging',
    ):
        """Make all nodes perform ``merging`` until none are busy.
        
        Emergent effect: merge touching clusters, push defects to roots.
        
        
        :param whole: whether to perform ``MERGING_WHOLE`` or ``MERGING_HALF`` stage.
        :param log_history: as in ``decode`` inputs.
        :param time_only: as in ``decode`` inputs.
        
        Output: ``t`` number of timesteps to complete growth round.
        """
        t = -1 if time_only == 'merging' else 1 if time_only == 'all' else 0

        while True:
            for node in self.NODES.values():
                node.merging(whole)
            for node in self.NODES.values():
                node.update_after_merging()
            t += (time_only!='unrooting') or any(node.cid==RESET for node in self.NODES.values())
            if not any(node.busy for node in self.NODES.values()):
                break
            if log_history == 'fine': self.append_history()

        if log_history == 'coarse': self.append_history()
        return t

    # DRAWERS

    def _labels(
            self,
            show_global=True,
            show_2_1_schedule_variables=True,
    ):
        """Return the labels dictionary for the drawer.
        
        
        :param show_global: whether to prepend the global label to the top-left node label.
        :param show_2_1_schedule_variables: whether to show
            node variables specific to the 2:1 cluster growth schedule.
        
        
        :returns: ``result`` a dictionary where each key a node index as a tuple; value, the label for the node at that index.
        """
        result = {v: node.label(show_2_1_schedule_variables)
                  for v, node in self.NODES.items()}
        if show_global:
            t = self.CODE.SCHEME.WINDOW_HEIGHT - 1
            top_left = (0, -1, t) if self.CODE.DIMENSION == 3 else (-1, t)
            result[top_left] = str(self._stage) + result[top_left]
        return result

    @override
    def draw_growth(
        self,
        highlighted_edges: set[Edge] | None = None,
        highlighted_edge_color='k',
        unhighlighted_edge_color=constants.DARK_GRAY,
        x_offset=constants.STREAM_X_OFFSET,
        with_labels=True,
        labels: dict[Node, str] | None = None,
        show_global=True,
        show_2_1_schedule_variables: None | bool = None,
        node_size: int | None = None,
        linewidths: float | None = None,
        active_shape='s',
        width: float | None = None,
        arrows: bool | None = None,
        show_boundary_defects=True,
        black_and_white=False,
        # following 2 kwargs are only for `black_and_white = True`
        bw_unhighlighted_width=constants.MEDIUM_THIN,
        bw_unrooted_color=constants.GRAY,
        **kwargs_for_networkx_draw,
    ):
        g = self.CODE.GRAPH
        dig, dig_diedges, dig_edges = self._pointer_digraph
        pos = self.CODE.get_pos(x_offset)
        unrooted_nodes = {v for v, node in self.NODES.items() if node.unrooted}
        if highlighted_edges is None:
            highlighted_edges = self.correction
        if show_2_1_schedule_variables is None:
            show_2_1_schedule_variables = isinstance(self._SCHEDULE, _TwoOne)
        if labels is None:
            labels = self._labels(
                show_global=show_global,
                show_2_1_schedule_variables=show_2_1_schedule_variables,
            )
        if node_size is None:
            node_size = self.BW_DEFAULT_NODE_SIZE if black_and_white else constants.DEFAULT
        if linewidths is None:
            linewidths = constants.THIN if black_and_white else constants.MEDIUM
        if width is None:
            width = constants.WIDE if black_and_white else constants.WIDE_MEDIUM

        if black_and_white:
            return self._draw_growth_black_and_white(
                g, dig, dig_diedges, dig_edges,
                pos, unrooted_nodes,
                highlighted_edges, with_labels, labels,
                node_size=node_size,
                unrooted_color=bw_unrooted_color,
                linewidths=linewidths,
                active_shape=active_shape,
                width=width,
                arrows=arrows,
                unhighlighted_width=bw_unhighlighted_width,
                show_boundary_defects=show_boundary_defects,
                **kwargs_for_networkx_draw,
            )

        # DRAW INACTIVE NODES AND EDGES UNUSED BY POINTERS
        
        # node-related kwargs
        inactive_nodelist = [v for v, node in self.NODES.items() if not node.active]
        inactive_node_color, inactive_edgecolors = self._get_node_color_and_edgecolors(
            outlined_nodes=unrooted_nodes,
            nodelist=inactive_nodelist,
            show_boundary_defects=show_boundary_defects,
        )
        
        # edge-related kwargs
        edgelist = [
            e for e, edge in self.EDGES.items()
            if edge.growth in {Growth.HALF, Growth.FULL}
            and e not in dig_edges
        ]
        edge_color, style = self._get_edge_color_and_style(
            edgelist,
            highlighted_edges,
            highlighted_edge_color,
            unhighlighted_edge_color,
        )
        nx.draw(
            g,
            pos=pos,
            with_labels=with_labels,
            labels=labels,
            nodelist=inactive_nodelist,
            node_size=node_size,
            node_color=inactive_node_color,
            linewidths=linewidths,
            edgecolors=inactive_edgecolors,
            edgelist=edgelist,
            width=width,
            edge_color=edge_color,
            style=style,
            **kwargs_for_networkx_draw,
        )

        # DRAW ACTIVE NODES
        active_nodelist = [v for v, node in self.NODES.items() if node.active]
        active_node_color, active_edgecolors = self._get_node_color_and_edgecolors(
            outlined_nodes=unrooted_nodes,
            nodelist=active_nodelist,
            show_boundary_defects=show_boundary_defects,
        )
        # need not actually pass `show_boundary_defects`
        # as boundary nodes are never active
        # but include in case need test correctness
        nx.draw_networkx_nodes(
            g,
            pos,
            nodelist=active_nodelist,
            node_size=node_size,
            node_color=active_node_color, # type: ignore
            node_shape=active_shape,
            linewidths=linewidths,
            edgecolors=active_edgecolors,
        )

        # DRAW EDGES USED BY POINTERS
        dig_edge_color = [
            highlighted_edge_color if e in highlighted_edges else
            unhighlighted_edge_color for e in dig_edges
        ]
        nx.draw_networkx_edges(
            dig,
            pos,
            node_size=node_size,
            edgelist=dig_diedges,
            width=width,
            edge_color=dig_edge_color, # type: ignore
            arrows=arrows,
        )

    def _draw_growth_black_and_white(
        self,
        g: nx.Graph, dig: nx.DiGraph, dig_diedges: list[Edge], dig_edges: list[Edge],
        pos: dict[Node, Coord], unrooted_nodes: set[Node],
        highlighted_edges: set[Edge], with_labels: bool, labels: dict[Node, str],
        node_size=BW_DEFAULT_NODE_SIZE,
        unrooted_color=constants.GRAY,
        linewidths=constants.THIN,
        active_shape='s',
        width=constants.WIDE,
        unhighlighted_width=constants.MEDIUM_THIN,
        arrows: bool | None = None,
        show_boundary_defects=True,
        **kwargs_for_networkx_draw,
    ):

        # DRAW INACTIVE DETECTORS AND EDGES UNUSED BY POINTERS
        
        # node-related kwargs
        inactive_detector_list = [v for v, node in self.NODES.items() if not (node.active or node._IS_BOUNDARY)]
        inactive_detector_linewidths = [width if v in self.verbose_syndrome else linewidths for v in inactive_detector_list]
        inactive_detector_color = [unrooted_color if v in unrooted_nodes else 'w' for v in inactive_detector_list]
        
        # edge-related kwargs
        edgelist = [
            e for e, edge in self.EDGES.items()
            if edge.growth in {Growth.HALF, Growth.FULL}
            and e not in dig_edges
        ]
        edge_width = [
            width if e in highlighted_edges else
            unhighlighted_width for e in edgelist
        ]
        style = [
            ':' if self.growth[e] is Growth.HALF else
            '-' for e in edgelist
        ]
        nx.draw(
            g,
            pos=pos,
            with_labels=with_labels,
            labels=labels,
            nodelist=inactive_detector_list,
            node_size=node_size,
            node_color=inactive_detector_color,
            linewidths=inactive_detector_linewidths,
            edgecolors='k',
            edgelist=edgelist,
            width=edge_width,
            edge_color='k',
            style=style,
            **kwargs_for_networkx_draw,
        )

        # DRAW ACTIVE DETECTORS
        active_detector_list = [v for v, node in self.NODES.items() if node.active]
        active_detector_linewidths = [width if v in self.verbose_syndrome else linewidths for v in active_detector_list]
        active_detector_color = [unrooted_color if v in unrooted_nodes else 'w' for v in active_detector_list]
        nx.draw_networkx_nodes(
            g,
            pos,
            nodelist=active_detector_list,
            node_size=node_size,
            node_color=active_detector_color, # type: ignore
            node_shape=active_shape,
            linewidths=active_detector_linewidths,
            edgecolors='k',
        )

        # DRAW BOUNDARY NODES
        boundary_nodelist = [v for v, node in self.NODES.items() if node._IS_BOUNDARY]
        boundary_linewidths = [width if v in self.verbose_syndrome else linewidths for v in boundary_nodelist] \
            if show_boundary_defects else linewidths
        nx.draw_networkx_nodes(
            g,
            pos,
            nodelist=boundary_nodelist,
            node_size=node_size,
            node_color='w',
            node_shape='8',
            linewidths=boundary_linewidths,
            edgecolors='k',
        )

        # DRAW EDGES USED BY POINTERS
        dig_edge_width = [
            width if e in highlighted_edges else
            unhighlighted_width for e in dig_edges
        ]
        nx.draw_networkx_edges(
            dig,
            pos,
            node_size=node_size,
            edgelist=dig_diedges,
            width=dig_edge_width, # type: ignore
            edge_color='k',
            arrows=arrows,
        )

    @property
    def _pointer_digraph(self):
        """
        :return dig: A NetworkX digraph representing the fully grown edges used by pointers.
        :return dig_diedges: The set of the digraph's edges as directed edges.
        :return dig_edges: The set of the digraph's edges as undirected edges.
        """
        # TODO: this is a temporary fix of ``_DigraphMaker.pointer_digraph``.
        # i.e. ``self.__init__`` used to have the line
        # ``self._DIGRAPH_MAKER = DigraphMaker(self.NODES, self.growth)``.
        dig = nx.DiGraph()
        dig.add_nodes_from(self.NODES.keys())
        dig_diedges: list[Edge] = []
        dig_edges: list[Edge] = []
        for u, node in self.NODES.items():
            if node.pointer != 'C':
                try:
                    e, index = node.NEIGHBORS[node.pointer]
                    if self.EDGES[e].growth is Growth.FULL:
                        v = e[index]
                        dig.add_edge(u, v)
                        dig_diedges.append((u, v))
                        dig_edges.append(e)
                except KeyError: pass  # node in bottom sheet points down
        return dig, dig_diedges, dig_edges
    
    @override
    def draw_decode(self, **kwargs_for_networkx_draw):
        self._DECODE_DRAWER.draw(self.history, **kwargs_for_networkx_draw)

    @property
    def _FIG_WIDTH(self):
        return max(1, self.CODE.D * self._FIG_FACTOR)
    
    @property
    def _FIG_HEIGHT(self):
        return max(1, (self.CODE.SCHEME.WINDOW_HEIGHT-1) * self._FIG_FACTOR)

    @property
    def _FIG_FACTOR(self):
        n = self.CODE.DIMENSION
        return 3*n*(n-1) / 10
    

    def generate_output(
            self,
            syndromes: list[str] | list[set[Node]],
            output_to_csv_file: str | None = None,
            draw: Literal[False, 'fine', 'coarse'] = False,
            margins=(0.2, 0.2),
            style: Literal['interactive', 'horizontal', 'vertical'] = 'interactive',
            **kwargs_for_draw_decode,
    ):
        """Generate output in the form of bitstrings.
        
        
        :param syndromes: the input of Snowflake.
            This can either be a list of sets of coordinates (each set representing a syndrome;
            each coordinate specifying a defect), or a list of syndrome vectors (strings of '0's
            and '1's). The vertical (i.e. time) coordinate of each defect must be
            ``self.CODE.SCHEME.WINDOW_HEIGHT-1``.

            For syndrome vectors, the ordering of the nodes is:

            - repetition code: west to east.
            - surface code: west to east along each row, then from south to north.

        :param output_to_csv_file: the CSV file path to save the data in e.g. 'snowflake_data.csv'.
            Defaults to ``None``, meaning no CSV file is saved.
        :param draw: whether to skip drawing the decoding process,
            draw it finely or draw it coarsely.
        :param margins: margins for the drawing.
        :param style: how different drawing frames are laid out.
            Can be 'interactive', 'horizontal', or 'vertical'.
        :param kwargs_for_draw_decode: additional keyword arguments for ``decoder.draw_decode()``.
        
        
        :returns: The output of Snowflake. This is a list of strings, each representing the edges in the bottom layer that are flipped just before each drop. The ordering of the edges is given by ``self._LOWEST_EDGES``. INCONSISTENCY: for the repetition code, the purely timelike edges in the last layer are excluded from ``self._LOWEST_EDGES``; the spacelike edges are ordered from west to east.
        
        Side effects:
        
        * If ``output_to_csv_file`` is not None, the input and output of Snowflake
            are saved in a CSV file at path ``output_to_csv_file``.
        * If ``draw`` is True, the decoding process is drawn.
        """
        first_syndrome, *_ = syndromes
        if type(first_syndrome) is str:
            # pad `syndrome_vectors` with zero vectors
            syndrome_vectors: list[str] = syndromes + self.CODE.SCHEME.WINDOW_HEIGHT * ['0'*self._BITSTRING_CONVERTER.LENGTH] # type: ignore
            syndrome_sets = self._BITSTRING_CONVERTER.syndrome_vectors_to_sets(syndrome_vectors)
        else:
            # pad `syndrome_sets` with empty sets
            syndrome_sets: list[set[Node]] = syndromes + self.CODE.SCHEME.WINDOW_HEIGHT * [set()] # type: ignore
            syndrome_vectors = self._BITSTRING_CONVERTER.sets_to_syndrome_vectors(syndrome_sets)
        
        self.init_history()
        for syndrome in syndrome_sets:
            self.decode(syndrome, log_history=draw, log_floor_history=True)
        if draw:
            self.draw_decode(
                margins=margins,
                style=style,
                **kwargs_for_draw_decode,
            )
        if output_to_csv_file is not None:
            import csv
            # Save syndrome vectors and corresponding floor history to CSV
            with open(output_to_csv_file, 'w', newline='') as csvfile:
                writer = csv.writer(csvfile)
                # Write header
                writer.writerow(['syndrome_vector', 'floor_history'])
                # Write data rows
                for syndrome_vector, floor_vector in zip(syndrome_vectors, self.floor_history, strict=True):
                    writer.writerow([syndrome_vector, floor_vector])
        return self.floor_history


class _BitstringConverter(abc.ABC):
    """Converts syndromes from bitstring format to sets of coordinates format."""

    @property
    @abc.abstractmethod
    def LENGTH(self):
        """The expected length of the syndrome vector."""

    def __init__(self, d: int, window_height: int):
        """
        :param d: the code distance.
        :param window_height: the height of the viewing window.
        """
        self._D = d
        self._WINDOW_HEIGHT = window_height

    def syndrome_vectors_to_sets(
            self,
            vectors: Sequence[str],
        ) -> list[set[Node]]:
        """Convert a sequence of syndrome vectors to a list of sets of defects.
        
        
        :param vectors: a sequence of syndrome vectors, each a string of '0's and '1's.
        
        
        :returns: A list of sets of defects, each set corresponding to a syndrome vector.
        """
        return [self._syndrome_vector_to_set(vector) for vector in vectors]

    @abc.abstractmethod
    def _syndrome_vector_to_set(
            self,
            vector: str,
        ) -> set[Node]:
        """Convert a syndrome vector to a set of defects.
        
        The ordering of the nodes is stated in ``Snowflake.generate_output`` docstring.
        """

    def sets_to_syndrome_vectors(
            self,
            syndromes: Sequence[set[Node]],
        ) -> list[str]:
        """Convert a sequence of sets of defects to a list of syndrome vectors.
        
        
        :param syndromes: a sequence of sets of defects.
        
        
        :returns: A list of syndrome vectors, each a string of '0's and '1's, each corresponding to a set of defects.
        """
        return [self._set_to_syndrome_vector(syndrome) for syndrome in syndromes]

    @abc.abstractmethod
    def _set_to_syndrome_vector(
            self,
            syndrome: set[Node],
    ) -> str:
        """Convert a set of defects to a syndrome vector.
        
        The ordering of the nodes is stated in ``Snowflake.generate_output`` docstring.
        """

    @abc.abstractmethod
    def lowest_edges(
            self,
            noise_model: str,
            _include_timelike_lowest_edges: bool = True,
    ) -> tuple[Edge, ...]:
        """Return the edges in the bottom layer of the viewing window
        in the order given by ``Snowflake._LOWEST_EDGES``.
        
        
        :param noise_model: the noise model of the code.
        :param _include_timelike_lowest_edges: whether to include the purely timelike edges.
        """


class _Repetition(_BitstringConverter):

    @property
    def LENGTH(self):
        return self._D - 1
    
    def _syndrome_vector_to_set(self, vector):
        if len(vector) != self.LENGTH:
            raise ValueError(f'Repetition code of distance {self._D} expects syndrome vector of length {self.LENGTH} but received length {len(vector)}.')
        return {(j, self._WINDOW_HEIGHT-1) for j, bit in enumerate(vector) if bit == '1'}
    
    def _set_to_syndrome_vector(self, syndrome):
        ones = {j for j, _ in syndrome}
        return ''.join('1' if j in ones else '0' for j in range(self.LENGTH))
    
    def lowest_edges(self, noise_model, _include_timelike_lowest_edges=True):
        up = tuple(((j, 0), (j, 1)) for j in range(self._D-1)) \
            if _include_timelike_lowest_edges else ()
        east = tuple(((j, 0), (j+1, 0)) for j in range(-1, self._D-1))
        if noise_model == 'phenomenological':
            return up + east
        else:
            raise NotImplementedError(f"Noise model {noise_model} not implemented for repetition code.")


class _Surface(_BitstringConverter):

    @property
    def LENGTH(self):
        return self._D * (self._D - 1)

    def _syndrome_vector_to_set(self, vector):
        if len(vector) != self.LENGTH:
            raise ValueError(f'Surface code of distance {self._D} expects syndrome vector of length {self.LENGTH} but received length {len(vector)}.')
        return {self._bit_position_to_defect(position) for position, bit in enumerate(vector) if bit == '1'}
    
    def _bit_position_to_defect(self, position: int) -> Node:
        """Convert bit position ``position`` to defect coordinate (i, j, t)."""
        return (
            self._D-1 - (position // (self._D-1)),
            position % (self._D-1),
            self._WINDOW_HEIGHT-1,
        )
    
    def _set_to_syndrome_vector(self, syndrome):
        ones = {self._defect_to_bit_position(defect) for defect in syndrome}
        return ''.join('1' if position in ones else '0' for position in range(self.LENGTH))
    
    def _defect_to_bit_position(self, defect: Node) -> int:
        """Convert defect coordinate (i, j, t) to a bit position."""
        i, j, t = defect
        if not (0 <= i < self._D and 0 <= j < self._D-1 and t == self._WINDOW_HEIGHT-1):
            raise ValueError(f'Defect {defect} out of range for surface code of distance {self._D} with window height {self._WINDOW_HEIGHT}.')
        return (self._D-1 - i) * (self._D-1) + j
    
    def lowest_edges(self, noise_model, _include_timelike_lowest_edges=True):
        d = self._D
        up = tuple(
            ((i, j, 0), (i, j, 1)) for i in range(d) for j in range(d-1)) \
            if _include_timelike_lowest_edges else ()
        south_down = tuple(
            ((i, j, 1), (i+1, j, 0)) for i in range(d-1) for j in range(d-1))
        east_up = tuple(
            ((i, j, 0), (i, j+1, 1)) for i in range(d) for j in range(d-2))
        south_east_up = tuple(
            ((i, j, 0), (i+1, j+1, 1)) for i in range(d-1) for j in range(d-2))
        south = tuple(
            ((i, j, 0), (i+1, j, 0)) for i in range(d-1) for j in range(d-1))
        east = tuple(
            ((i, j, 0), (i, j+1, 0)) for i in range(d) for j in range(-1, d-1))
        if noise_model == 'phenomenological':
            return up + south + east
        elif noise_model == 'circuit-level':
            return up + south_down + east_up + south_east_up + south + east
        else:
            raise NotImplementedError(f"Noise model {noise_model} not implemented for surface code.")


class NodeEdgeMixin(abc.ABC):
    """Mixin class for ``_Node`` and ``_Edge``.
    
    Instance attributes:
    
    * ``SNOWFLAKE`` the decoder the node or edge belongs to.
    """

    _SNOWFLAKE: Snowflake

    @property
    def SNOWFLAKE(self): return self._SNOWFLAKE


class _Node(NodeEdgeMixin):
    """Node for Snowflake.
    
    Extends ``NodeEdgeMixin``.
    
    Additional instance attributes:
    
    * ``INDEX`` index of node.
    * ``ID`` the unique ID of node.
    * ``FRIENDSHIP`` the type of connection the node has for communicating.
    * ``NEIGHBORS`` a dictionary where each
        key a pointer string;
        value, a tuple of the form (edge, index of edge which is neighbor).
    * ``_IS_BOUNDARY`` whether node is a boundary node.
    * ``_MERGER`` the provider for the ``merging`` method.
    * ``UNROOTER`` the type of unrooting process used.
    * ``active, whole, cid, defect, pointer, grown, unrooted, busy``
        explained in [arXiv:2406.01701, C.1].
    * The variables in the above bulletpoint, save ``busy``,
        each have ``next_`` versions for next timestep.
    * ``access`` refers to neighbors along fully grown edges.
        In the form of a dictionary of where each
        key a pointer string;
        value, the ``_Node`` object in the direction of that pointer.
    """

    def __init__(
            self,
            snowflake: Snowflake,
            index: Node,
            merger: Literal['fast', 'slow'] = 'fast',
            unrooter: Literal['full', 'simple'] = 'full',
        ) -> None:
        """
        :param snowflake: the decoder the node belongs to.
        :param index: index of node.
        :param merger: decides whether to flood before syncing (fast) or vice versa (slow) in a merging step.
        :param unrooter: the type of unrooting process to use.
        """
        self._SNOWFLAKE = snowflake
        self._INDEX = index
        self._ID = snowflake.index_to_id(index)
        self._FRIENDSHIP = NothingFriendship(self) if index[snowflake.CODE.TIME_AXIS] == 0 \
            else TopSheetFriendship(self) if index[snowflake.CODE.TIME_AXIS] == snowflake.CODE.SCHEME.WINDOW_HEIGHT-1 \
            else NodeFriendship(self)
        if isinstance(snowflake.CODE, Repetition):
            j, t = index
            provisional_neighbors: dict[direction, tuple[Edge, int]] = {
                'W': (((j-1, t), index), 0),
                'E': ((index, (j+1, t)), 1),
                'D': (((j, t-1), index), 0),
                'U': ((index, (j, t+1)), 1),
            }
        else:  # Surface
            i, j, t = index
            provisional_neighbors: dict[direction, tuple[Edge, int]] = {
                'NWD': (((i-1, j-1, t-1), index), 0),
                'N': (((i-1, j, t), index), 0),
                'NU': (((i-1, j, t+1), index), 0),
                'WD': (((i, j-1, t-1), index), 0),
                'W': (((i, j-1, t), index), 0),
                'D': (((i, j, t-1), index), 0),
                'U': ((index, (i, j, t+1)), 1),
                'E': ((index, (i, j+1, t)), 1),
                'EU': ((index, (i, j+1, t+1)), 1),
                'SD': ((index, (i+1, j, t-1)), 1),
                'S': ((index, (i+1, j, t)), 1),
                'SEU': ((index, (i+1, j+1, t+1)), 1),
            }
        self._NEIGHBORS: dict[direction, tuple[Edge, int]] = {
            pointer: provisional_neighbors[pointer]
            for pointer in snowflake._NEIGHBOR_ORDER
            if provisional_neighbors[pointer][0] in snowflake.EDGES
        }
        self._IS_BOUNDARY = snowflake.CODE.is_boundary(index)
        self._MERGER = _FastMerger(self) if merger == 'fast' else _SlowMerger(self)
        self._UNROOTER = _FullUnrooter(self) if unrooter == 'full' else _SimpleUnrooter(self)
        self.reset()

    def __repr__(self) -> str:
        return f'decoders.snowflake._Node({self._SNOWFLAKE}, {self.INDEX})'
    
    def __str__(self) -> str:
        return str(self.INDEX)

    @property
    def INDEX(self): return self._INDEX

    @property
    def ID(self): return self._ID

    @property
    def FRIENDSHIP(self): return self._FRIENDSHIP

    @property
    def NEIGHBORS(self): return self._NEIGHBORS

    @property
    def UNROOTER(self): return self._UNROOTER

    def label(self, show_2_1_schedule_variables=True):
        cid = str(self.cid) if self.cid != RESET else 'R'
        if show_2_1_schedule_variables:
            if self.grown:
                grown_half = "'" if self.whole else ":"
            else:
                grown_half = "" if self.whole else "."
        else:
            grown_half = ""
        return cid + grown_half

    def reset(self):
        """Factory reset."""

        self.active = False
        self.whole = True
        self.cid = self.ID
        self.defect = False
        self.pointer: direction = 'C'
        self.grown = False
        self.unrooted = False

        self.next_active = False
        self.next_whole = True
        self.next_cid = self.ID
        self.next_defect = False
        self.next_pointer: direction = 'C'
        self.next_grown = False
        self.next_unrooted = False

        self.busy = False
        self.access: dict[direction, _Node] = {}

    def update_after_drop(self):
        """Update unphysicals after a drop.
        
        Set ``[attribute]`` := ``next_[attribute]``.
        For nodes in top sheet of viewing window, this equals ``reset``.
        """
        self.defect = self.next_defect
        self.active = self.next_active
        self.cid = self.next_cid
        self.pointer = self.next_pointer
        self.whole = self.next_whole
        # NEED NOT RESET...
        # `next_defect` as value correct for next use in `syncing`
        # `next_active` as always overwritten in `syncing` before next used in `update_after_merging`
        # `next_cid` as value correct for next use in `flooding`
        # `next_pointer` as always overwritten in `drop` before next used in `update_after_drop`
        # `next_whole` as always overwritten in `drop` before next used in `update_after_drop`

    def grow(self):
        """Call ``grow`` if active, then find broken pointers.
        
        Used only in 1:1 schedule.
        """
        if self.active:
            self._grow()
        self.FRIENDSHIP.find_broken_pointers()

    def grow_whole(self):
        """Call ``grow`` if active and whole, then find broken pointers."""
        if self.active and self.whole:
            self._grow()
            self.grown = True
            self.next_grown = True
        self.FRIENDSHIP.find_broken_pointers()

    def grow_half(self):
        """Call ``grow`` if active and half and not yet grown in current decoding cycle."""
        self.unrooted = False
        self.next_unrooted = False
        if self.active and not self.whole and not self.grown:
            self._grow()
    
    def _grow(self):
        """Growth-increment incident edges."""
        s = self.SNOWFLAKE
        edges_to_grow = {
            e for e, _ in self.NEIGHBORS.values()
            if s.EDGES[e].growth in s._ACTIVE_GROWTH_VALUES
        }
        for e in edges_to_grow:
            s.EDGES[e].growth += Growth.INCREMENT
        self.whole ^= True

    def update_access(self):
        """Update access.
        
        Call after growth has changed. This method is unphysical.
        """
        self.access = {
            pointer: self.SNOWFLAKE.NODES[e[index]]
            for pointer, (e, index) in self.NEIGHBORS.items()
            if self.SNOWFLAKE.EDGES[e].growth is Growth.FULL
        }

    def merging(self, whole: bool):
        """Advance 1 merging timestep.
        
        The emergent effect of each node running this method repeatedly
        is the merging of clusters.
        """
        self.busy = False
        self._MERGER.merging(whole)

    def syncing(self):
        """Update ``active`` depending on, and push defect to, pointee."""
        detector_defect = not self._IS_BOUNDARY and self.defect
        if self.pointer == 'C':
            self.next_active = detector_defect
        else:
            self.next_active = self.access[self.pointer].active
            if detector_defect:  # PUSH DEFECT
                self.busy = True
                # relay along pointer
                self.access[self.pointer].next_defect ^= True
                self.next_defect ^= True
                # flip edge used by pointer
                e, _ = self.NEIGHBORS[self.pointer]
                self.SNOWFLAKE.EDGES[e].correction ^= True
        if self.active != self.next_active:
            self.busy = True

    def flooding(self, whole: bool):
        """Update ``pointer, cid, unrooted, grown`` depending on access."""
        if whole:
            self.UNROOTER.flooding_whole()
        else:
            self.UNROOTER.flooding_half()

    def update_after_merging(self):
        """``next_{cid, defect, active, unrooted, grown}`` -> ``{cid, defect, active, unrooted, grown}``."""
        self.cid = self.next_cid
        self.defect = self.next_defect
        self.active = self.next_active
        self.unrooted = self.next_unrooted
        self.grown = self.next_grown
        # NEED NOT RESET...
        # `next_cid` as value correct for next use in `flooding`
        # `next_defect` as value correct for next use in `syncing`
        # `next_active` as always overwritten in `syncing` before next used in `update_after_merging`
        # `next_unrooted` as value correct (unless edited) for next use in `update_after_merging`
        # `next_grown` as value correct for next use in `flooding`


class Friendship(abc.ABC):
    """The type of connection for communicating
    ``defect, active, cid, pointer`` information.
    
    Instance attributes (1 constant):
    
    * ``NODE`` the node which has this friendship.
    """

    def __init__(self, node: _Node) -> None:
        self._NODE = node

    @property
    def NODE(self): return self._NODE

    def drop(self):
        """Drop information to node immediately below."""
        self.NODE.grown = False
        self.NODE.next_grown = False
        self.NODE.unrooted = False
        self.NODE.next_unrooted = False

    def find_broken_pointers(self):
        """Start unrooting if in bottom sheet of viewing window and point downward."""


class NodeFriendship(Friendship):
    """Friendship with node immediately below.
    
    Extends ``Friendship``.
    
    Instance attributes (1 constant):
    
    * ``DROPEE`` the node immediately below.
    """

    def __init__(self, node: _Node) -> None:
        self._DROPEE = node.SNOWFLAKE.CODE.raise_node(
            node.INDEX,
            delta_t=-1,
        )
        super().__init__(node)

    @property
    def DROPEE(self): return self._DROPEE

    def drop(self):
        super().drop()
        r = self.NODE.SNOWFLAKE.NODES[self.DROPEE]
        r.next_defect = self.NODE.defect
        r.next_active = self.NODE.active
        r.next_cid = r.SNOWFLAKE.id_below(self.NODE.cid)
        r.next_pointer = self.NODE.pointer
        r.next_whole = self.NODE.whole

class TopSheetFriendship(NodeFriendship):
    """Friendship for nodes in top sheet of viewing window.
    
    Extends ``NodeFriendship``.
    After a drop, these nodes must reset ``next_active`` and ``next_defect``.
    """

    def drop(self):
        super().drop()
        self.NODE.next_defect = False
        self.NODE.next_active = False


class NothingFriendship(Friendship):
    """Friendship with nothing immediately below.
    
    Extends ``Friendship``.
    Only nodes whose time index is 0 have this friendship.
    """

    def find_broken_pointers(self):
        if 'D' in self.NODE.pointer:
            self.NODE.UNROOTER.start()


class _Merger(abc.ABC):
    """The class providing ``_Node.merging``."""

    def __init__(self, node: _Node) -> None:
        self._NODE = node

    @abc.abstractmethod
    def merging(self):
        """See ``_Node.merging``."""


class _SlowMerger(_Merger):
    """Provider for ``_Node.merging`` which syncs before it floods.
    
    Extends ``_Merger``.
    """

    def merging(self, whole: bool):
        self._NODE.syncing()
        self._NODE.flooding(whole)


class _FastMerger(_Merger):
    """Provider for ``_Node.merging`` which floods before it syncs.
    
    Extends ``_Merger``.
    """

    def merging(self, whole: bool):
        self._NODE.flooding(whole)
        self._NODE.syncing()


class _Schedule(abc.ABC):
    """Abstract base class for the cluster growth schedule."""

    def __init__(self, snowflake: Snowflake) -> None:
        self._SNOWFLAKE = snowflake

    @abc.abstractmethod
    def finish_decode(
        self,
        log_history: Literal[False, 'fine', 'coarse'] = False,
        time_only: Literal['all', 'merging', 'unrooting'] = 'merging',
    ) -> int:
        """Perform the rest of the decoding cycle after drop.
        
        
        :param log_history: As in ``decode`` inputs.
        :param time_only: As in ``decode`` inputs.
        
        
        :return t: Number of timesteps to complete decoding cycle.
        """

    @abc.abstractmethod
    def grow(self):
        """Make all nodes perform a ``grow``."""


class _OneOne(_Schedule):
    """The 1:1 cluster growth schedule.
    
    Extends ``_Schedule``.
    """

    def finish_decode(self, log_history, time_only):
        self._SNOWFLAKE._stage += Stage.INCREMENT
        self.grow()
        if log_history == 'fine': self._SNOWFLAKE.append_history()
        self._SNOWFLAKE._stage += Stage.INCREMENT
        return self._SNOWFLAKE.merge(
            whole=True,
            log_history=log_history,
            time_only=time_only,
        )
    
    def grow(self):
        for node in self._SNOWFLAKE.NODES.values():
            node.grow()
        for node in self._SNOWFLAKE.NODES.values():
            node.update_access()  # TODO: speed up by changing just the accesses affected


class _TwoOne(_Schedule):
    """The 2:1 cluster growth schedule.
    
    Extends ``_Schedule``.
    """

    def finish_decode(self, log_history, time_only):
        t = 0
        for whole in (True, False):
            self._SNOWFLAKE._stage += Stage.INCREMENT
            self.grow(whole)
            if log_history == 'fine': self._SNOWFLAKE.append_history()
            self._SNOWFLAKE._stage += Stage.INCREMENT
            t += self._SNOWFLAKE.merge(
                whole,
                log_history,
                time_only=time_only,
            )
        return t
    
    def grow(self, whole: bool):
        if whole:
            for node in self._SNOWFLAKE.NODES.values():
                node.grow_whole()
        else:
            for node in self._SNOWFLAKE.NODES.values():
                node.grow_half()
        for node in self._SNOWFLAKE.NODES.values():
            node.update_access()  # TODO: speed up by changing just the accesses affected


class _Unrooter(abc.ABC):
    """Abstract base class for the type of unrooting process to use.
    
    Instance constants:
    
    * ``_NODE`` the node which has this unrooter.
    """

    def __init__(self, node: _Node):
        """Input: ``node`` the node the unrooter belongs to."""
        self._NODE = node

    @abc.abstractmethod
    def start(self):
        """Start unrooting the node."""

    @abc.abstractmethod
    def flooding_whole(self):
        """Update ``pointer, cid, unrooted, grown`` depending on access."""

    def flooding_half(self):
        """Update ``pointer, cid`` depending on access."""
        for pointer, neighbor in self._NODE.access.items():
            self._compare_cid(pointer, neighbor)

    def _compare_cid(self, pointer: direction, neighbor: _Node):
        """Update ``next_cid`` depending on ``neighbor.cid``."""
        if neighbor.cid < self._NODE.next_cid:
            self._NODE.busy = True
            self._NODE.pointer = pointer
            self._NODE.next_cid = neighbor.cid

    def _check_grown(self, neighbor: _Node):
        """Update ``next_grown`` depending on ``neighbor.grown``."""
        if neighbor.grown and not self._NODE.next_grown:
            self._NODE.busy = True
            self._NODE.next_grown = True


class _FullUnrooter(_Unrooter):
    """Full unrooting process where each node in the amputated cluster resets its CID and pointer.
    
    This is so that the pointer tree structure can be rebuilt from scratch,
    via further merging timesteps.
    
    Extends ``_Unrooter``.
    """

    def start(self):
        self._NODE.cid = RESET
        self._NODE.pointer = 'C'
        # NEED NOT RESET...
        # `next_cid` as always overwritten in `NODE.flooding`
        # `next_pointer` as always overwritten in `drop` before next used in `NODE.update_after_drop`

    def flooding_whole(self):
        if self._NODE.cid == RESET:  # finish unrooting `self._NODE`
            self._NODE.busy = True
            self._NODE.next_cid = self._NODE.ID
            self._NODE.next_unrooted = True
        else:
            for pointer, neighbor in self._NODE.access.items():
                if neighbor.cid == RESET:
                    if not self._NODE.unrooted:  # start unrooting `self._NODE`
                        self._NODE.busy = True
                        self._NODE.next_cid = RESET
                        self._NODE.pointer = 'C'
                        break
                else:
                    self._compare_cid(pointer, neighbor)
                self._check_grown(neighbor)


class _SimpleUnrooter(_Unrooter):
    """Simple unrooting process where the node at breaking point only establishes the shortest path to a boundary.
    
    Extends ``_Unrooter``.
    
    Additional instance constants:
    
    * ``_CLOSEST_BOUNDARY_DIRECTION`` the direction toward the closest boundary.
    """

    def __init__(self, node: _Node):
        super().__init__(node)
        code = node.SNOWFLAKE.CODE
        self._CLOSEST_BOUNDARY_DIRECTION = 'W' if node.INDEX[code.LONG_AXIS] < (code.D-1)/2 else 'E'

    def start(self):
        self._NODE.cid = RESET
        self._NODE.next_cid = RESET

    def flooding_whole(self):
        if self._NODE.cid == RESET:
            if not self._NODE.unrooted and not self._NODE._IS_BOUNDARY:
                self._wave()
        else:
            for pointer, neighbor in self._NODE.access.items():
                if neighbor.cid != RESET:
                    self._compare_cid(pointer, neighbor)
                self._check_grown(neighbor)

    def _wave(self):
        """Propagate unroot wave toward nearest boundary."""
        self._NODE.next_unrooted = True
        self._NODE.pointer = self._CLOSEST_BOUNDARY_DIRECTION
        e, index = self._NODE.NEIGHBORS[self._NODE.pointer]
        u = self._NODE.SNOWFLAKE.NODES[e[index]]
        self._NODE.SNOWFLAKE.EDGES[e].growth = Growth.FULL  # physical
        self._NODE.access[self._NODE.pointer] = u  # unphysical
        u.next_cid = RESET


class _Edge(NodeEdgeMixin):
    """Edge for Snowflake.
    
    Extends ``NodeEdgeMixin``.
    
    Additional instance attributes:
    
    * ``INDEX`` index of edge.
    * ``CONTACT`` the type of connection the edge has for communicating.
        Analogous to ``FRIENDSHIP`` for nodes.
    * ``growth`` its growth value.
    * ``correction`` whether it is in the correction.
    """

    def __init__(self, snowflake: Snowflake, index: Edge) -> None:
        """
        :param snowflake: the decoder the edge belongs to.
        :param INDEX: index of edge.
        """
        self._SNOWFLAKE = snowflake
        self._INDEX = index
        lowness = sum(v[snowflake.CODE.TIME_AXIS] == 0 for v in index)
        self._CONTACT = EdgeContact(self) if lowness == 0 else FloorContact(self)
        self.reset()

    def __repr__(self) -> str:
        return f'decoders.snowflake._Edge({self.SNOWFLAKE}, {self.INDEX})'
    
    def __str__(self) -> str:
        return str(self.INDEX)

    @property
    def INDEX(self): return self._INDEX

    @property
    def CONTACT(self): return self._CONTACT

    def reset(self):
        """Factory reset."""

        self.growth = Growth.UNGROWN
        self.correction = False

        self.next_growth = Growth.UNGROWN
        self.next_correction = False

    def update_after_drop(self):
        """Update unphysicals after a drop.
        
        Set ``[attribute]`` := ``next_[attribute]``.
        For edges in ``self.CODE.SCHEME.BUFFER_EDGES``,
        this equals ``reset``.
        """
        self.growth = self.next_growth
        self.correction = self.next_correction


class _Contact(abc.ABC):
    """The type of connection for communicating
    ``growth, correction`` information.
    
    Instance attributes (1 constant):
    
    * ``EDGE`` the edge which has this contact.
    """

    def __init__(self, edge: _Edge) -> None:
        self._EDGE = edge

    @property
    def EDGE(self): return self._EDGE

    @abc.abstractmethod
    def drop(self):
        """Drop information to edge or floor immediately below."""


class EdgeContact(_Contact):
    """Contact with edge immediately below.
    
    Extends ``_Contact``.
    
    Instance attributes (1 constant):
    
    * ``DROPEE`` the edge immediately below.
    """

    def __init__(self, edge: _Edge) -> None:
        code = edge.SNOWFLAKE.CODE
        self._DROPEE = code.raise_edge(edge.INDEX, delta_t=-1)
        super().__init__(edge)

    @property
    def DROPEE(self): return self._DROPEE

    def drop(self):
        r = self.EDGE.SNOWFLAKE.EDGES[self.DROPEE]
        r.next_growth = self.EDGE.growth
        r.next_correction = self.EDGE.correction


class FloorContact(_Contact):
    """Contact for edges in commit region.
    
    Extends ``_Contact``.
    Only edges in the bottom sheet of the viewing window
    have this contact.
    """

    def drop(self):
        frugal: Frugal = self.EDGE.SNOWFLAKE.CODE.SCHEME # type: ignore
        if self.EDGE.correction:
            frugal.pairs.load(self.EDGE.INDEX)


import pytest

from localuf._pairs import Pairs
from localuf.type_aliases import Node


@pytest.fixture
def pairs():
    return Pairs()


@pytest.fixture
def uvwx() -> tuple[Node, Node, Node, Node]:
    return (0, 0), (1, 0), (0, 1), (1, 1)


import pytest

from localuf.decoders.snowflake import _Edge

@pytest.mark.parametrize("prop", [
    "EDGE",
])
def test_property_attributes(test_property, sfe3: _Edge, prop):
    test_property(sfe3.CONTACT, prop)


import itertools

import pytest

from localuf import Surface
from localuf.decoders import Snowflake
from localuf._schemes import Frugal


@pytest.fixture
def sf_frugal3():
    sf3 = Surface(
        3,
        noise='phenomenological',
        scheme='frugal',
    )
    return sf3.SCHEME


def test_get_logical_error(sf_frugal3: Frugal):
    
    decoder = Snowflake(sf_frugal3._CODE)
    h = sf_frugal3.WINDOW_HEIGHT
    m = 0
    for error in itertools.chain(
        [
            {
                ((1, 1, h-1), (1, 1, h)),
                ((0, 1, h-1), (0, 1, h)),
            },
            {
                ((1, 1, h-1), (1, 2, h-1)),
            },
        ],
        itertools.repeat(set(), 2*h),
    ):
        sf_frugal3._raise_window()
        syndrome = sf_frugal3._load(error)
        decoder.decode(syndrome)
        m += sf_frugal3.get_logical_error()
    assert m == 0
    # below failed if pairs w/ zero pair separation (in j) are not added to `new_pairs`
    assert sf_frugal3.pairs._dc == {}


import pytest

from localuf import Repetition, Surface
from localuf.decoders import Snowflake
from localuf.decoders.snowflake import _Node, _Edge, Snowflake
from localuf.type_aliases import Node


@pytest.fixture(
        name="rp_frugal",
        params=range(3, 11, 2),
        ids=lambda x: f"d{x}",
)
def _rp_frugal(request):
    return Repetition(
        request.param,
        noise='phenomenological',
        scheme='frugal',
    )


@pytest.fixture
def snowflake(rp_frugal: Repetition):
    return Snowflake(rp_frugal)


@pytest.fixture
def snowflake_one_one(rp_frugal: Repetition):
    return Snowflake(rp_frugal, schedule='1:1')


@pytest.fixture
def frugal_rep_3():
    return Repetition(
        3,
        noise='phenomenological',
        scheme='frugal',
    )


@pytest.fixture
def snowflake3(frugal_rep_3: Repetition):
    return Snowflake(frugal_rep_3)


@pytest.fixture
def snowflake3_one_one():
    rp = Repetition(
        3,
        noise='phenomenological',
        scheme='frugal',
    )
    return Snowflake(rp, schedule='1:1')


@pytest.fixture
def sfn3(snowflake3: Snowflake):
    return snowflake3.NODES[0, 0]


@pytest.fixture
def fixture_test_INDEX_property():
    def f(obj: _Node | _Edge):
        assert type(obj.INDEX) is tuple
    return f


@pytest.fixture
def sfe3(snowflake3: Snowflake):
    return snowflake3.EDGES[(0, 0), (1, 0)]


@pytest.fixture
def syncing_flooding_objects(snowflake3: Snowflake) -> tuple[
    Snowflake, tuple[Node, Node, Node], tuple[_Node, _Node, _Node]
]:
    w, c, e = (-1, 0), (0, 0), (1, 0)
    west = snowflake3.NODES[w]
    center = snowflake3.NODES[c]
    east = snowflake3.NODES[e]
    return snowflake3, (w, c, e), (west, center, east)


@pytest.fixture
def surface3_CL_frugal():
    """Distance-3 surface code, circuit-level noise, frugal scheme."""
    return Surface(3, 'circuit-level', scheme='frugal')


from collections.abc import Mapping
from collections import defaultdict

import numpy as np
import numpy.typing as npt
from pymatching import Matching

from localuf._base_classes import Code, Decoder
from localuf._schemes import Batch
from localuf.type_aliases import Node, Edge


class MWPM(Decoder):
    """Minimum weight perfect matching decoder using PyMatching implementation.
    
    Extends ``Decoder``.
    Compatible so far only with the batch decoding scheme.
    """

    correction_vector: npt.NDArray[np.bool_]
    """The stored correction as a binary vector whose length is the number of fault IDs."""
    correction_weight: float = 0
    """The weight of the stored correction."""

    def __init__(self, code: Code, _edge_weight_modifier: None | dict[Edge, float] = None):
        """
        :param code: The code to be decoded.
        :param _edge_weight_modifier: Maps each edge to a multiplicative factor
            that scales the weight of that particular edge.
            Designed for implementing non-uniform weights for phenomenological noise model.
            If not specified, no scaling is done i.e. all edges have modifier 1.
        """
        if not isinstance(code.SCHEME, Batch):
            raise NotImplementedError
        super().__init__(code)
        self._DETECTOR_TO_INT = {v: k for k, v in enumerate(code.DETECTORS)}
        """Maps each detector node index to a unique integer in 0..detector_count."""
        self._WEST_LOGICAL: dict[Node, int | None] = {
            v: 0 if (v[code.LONG_AXIS] == -1) else None
            for v in code.BOUNDARY_NODES
        }
        """Maps the west (east) boundary node index to fault ID 0 (None)."""
        self._WEST_EAST_LOGICALS: dict[Node, int] = {
            v: 0 if (v[code.LONG_AXIS] == -1) else 1
            for v in code.BOUNDARY_NODES
        }
        """Maps the west (east) boundary node index to fault ID 0 (1)."""
        self._edge_weight_modifier: dict[Edge, float] = {
            e: 1.0 for e in code.EDGES
        } if _edge_weight_modifier is None else _edge_weight_modifier
        """Maps each edge to a multiplicative factor
        that scales the weight of that particular edge.
        Designed for implementing non-uniform weights for phenomenological noise model.
        """

    def get_matching(
            self,
            noise_level: None | float = None,
            detector_to_int: None | dict[Node, int] = None,
            boundary_node_to_fault_id: None | Mapping[Node, int | None] = None,
    ):
        """Return a PyMatching matching graph.
        
        
        :param noise_level: a probability that represents the noise strength.
            This is passed to ``self.CODE.NOISE.get_edge_weights()``.
            If ``None``, all edges in ``matching`` have error probability 0 and weight 1
            (before being scaled by ``self._edge_weight_modifier``).
        :param detector_to_int: maps each detector node index to a unique integer.
            If ``None``, use the default map from each index to a unique integer in 0..detector_count.
        :param boundary_node_to_fault_id: maps each boundary node index to its fault ID.
            If ``None``, use the default map from the west (east) boundary node index to fault ID 0 (None).
        
        :return matching: a PyMatching matching graph whose...
            1. edge weights and error probabilities depend on ``noise_level`` and ``self._edge_weight_modifier``,
            2. detectors are numbered according to ``detector_to_int``,
            3. boundary nodes are all virtual;
            all edges connected to each boundary node have the fault ID given by ``boundary_node_to_fault_id``.
        """
        if detector_to_int is None:
            detector_to_int = self._DETECTOR_TO_INT
        if boundary_node_to_fault_id is None:
            boundary_node_to_fault_id = self._WEST_LOGICAL
        edge_weights = self.CODE.NOISE.get_edge_weights(noise_level)
        matching = Matching()
        for (u, v), (flip_probability, weight) in edge_weights.items():
            modified_weight = self._edge_weight_modifier[u, v] * weight
            int_u = detector_to_int.get(u, None)
            int_v = detector_to_int.get(v, None)
            fault_u = boundary_node_to_fault_id.get(u, None)
            fault_v = boundary_node_to_fault_id.get(v, None)
            if int_u is None:
                # type(int_v) is int
                matching.add_boundary_edge(
                    node=int_v, # type: ignore
                    fault_ids=fault_u, # type: ignore
                    weight=modified_weight,
                    error_probability=flip_probability,
                )
            elif int_v is None:
                matching.add_boundary_edge(
                    node=int_u,
                    fault_ids=fault_v, # type: ignore
                    weight=modified_weight,
                    error_probability=flip_probability,
                )
            else:
                matching.add_edge(
                    node1=int_u,
                    node2=int_v,
                    weight=modified_weight,
                    error_probability=flip_probability,
                )
        return matching

    # unused
    def _get_matching_with_real_boundaries(
            self,
            noise_level: None | float = None,
            detector_to_int: None | dict[Node, int] = None,
            boundary_node_to_fault_id: None | Mapping[Node, int | None] = None,
    ):
        """Same as ``get_matching`` but the boundary nodes of the output are real
        and are numbered consecutively starting from ``detector_count``.
        """
        if detector_to_int is None:
            detector_to_int = self._DETECTOR_TO_INT
        if boundary_node_to_fault_id is None:
            boundary_node_to_fault_id = self._WEST_LOGICAL
        edge_weights = self.CODE.NOISE.get_edge_weights(noise_level)
        matching = Matching()
        # `boundary_nodes` maps each fault ID to a boundary node index
        boundary_nodes: defaultdict[int | None, int] = defaultdict(
            lambda: self.CODE.DETECTOR_COUNT + len(boundary_nodes))
        for (u, v), (flip_probability, weight) in edge_weights.items():
            int_u = detector_to_int.get(u, None)
            int_v = detector_to_int.get(v, None)
            fault_u = boundary_node_to_fault_id.get(u, None)
            fault_v = boundary_node_to_fault_id.get(v, None)
            if int_u is None:
                # type(int_v) is int
                matching.add_edge(
                    node1=boundary_nodes[fault_u],
                    node2=int_v, # type: ignore
                    fault_ids=fault_u, # type: ignore
                    weight=weight,
                    error_probability=flip_probability,
                )
            elif int_v is None:
                matching.add_edge(
                    node1=int_u,
                    node2=boundary_nodes[fault_v],
                    fault_ids=fault_v, # type: ignore
                    weight=weight,
                    error_probability=flip_probability,
                )
            else:
                matching.add_edge(
                    node1=int_u,
                    node2=int_v,
                    weight=weight,
                    error_probability=flip_probability,
                )
        matching.set_boundary_nodes(set(boundary_nodes.values()))
        return matching
    
    def reset(self):
        self.correction_vector = np.zeros(self.CODE.DETECTOR_COUNT, dtype=bool)
        self.correction_weight = 0

    def get_binary_vector(self, syndrome: set[Node], detector_to_int: None | dict[Node, int] = None):
        """Convert ``syndrome`` to a binary vector.
        
        
        :param syndrome: the set of defects.
        :param detector_to_int: maps each detector node index to a unique integer in 0..detector_count.
        
        
        :return binary_vector: A binary vector of length equal to detector count
            whose ordering is given by ``detector_to_int``.
            If not specified, uses ordering given by ``self.CODE.DETECTORS``.
        """
        if detector_to_int is None:
            detector_to_int = self._DETECTOR_TO_INT
        binary_vector: npt.NDArray[np.bool_] = np.zeros(self.CODE.DETECTOR_COUNT, dtype=bool)
        for v in syndrome:
            binary_vector[detector_to_int[v]] = 1
        return binary_vector
        
    def decode(self, syndrome, **kwargs):
        """Decode syndrome.
        
        
        :param syndrome: the set of defects.
        :param kwargs: passed to ``self.get_matching()``.
        
        Side effects:
        
        * Update ``self.correction_vector`` and ``self.correction_weight``.
        """
        syndrome_vector = self.get_binary_vector(syndrome)
        matching = self.get_matching(**kwargs)
        self.correction_vector, self.correction_weight = matching.decode(
            z=syndrome_vector,
            return_weight=True,
        )
    
    def draw_decode(self, **kwargs_for_networkx_draw):
        raise NotImplementedError
    
    def complementary_gap(
            self,
            syndrome: set[Node],
            noise_level: None | float = None,
    ) -> tuple[npt.NDArray[np.bool_], float]:
        """Decode ``syndrome`` and calculate the complementary gap.
        
        Does not require ``self.decode()`` to be called first.
        
        
        :param syndrome: the set of defects.
        :param noise_level: a probability representing the noise strength.
            This is needed to define nonuniform edge weights of the decoding graph
            in the circuit-level noise model.
            If ``None``, all edges are assumed to have weight 1.
        
        
        :return [correction_west, correction_east]: A 2-vector whose first (second) element
            is the parity of number of edges in the minimum-weight correction
            that connect to the west (east) boundary node.
        :return gap: The complementary gap.
        
        :raises ValueError: if ``self.CODE.MERGED_EQUIVALENT_BOUNDARY_NODES`` is ``False``.
        """
        if not self.CODE.MERGED_EQUIVALENT_BOUNDARY_NODES:
            raise ValueError('Complementary gap requires all equivalent boundary nodes in the decoding graph be merged.')
        syndrome_vector = self.get_binary_vector(syndrome)
        matching_1, matching_2 = self._get_complementary_gap_matchings(noise_level=noise_level)
        correction_1, weight_1 = matching_1.decode(syndrome_vector, return_weight=True)
        complementary_syndrome_vector = np.concatenate((syndrome_vector, correction_1 ^ True))
        _, weight_2 = matching_2.decode(complementary_syndrome_vector, return_weight=True)
        return correction_1, weight_2 - weight_1

    def _get_complementary_gap_matchings(self, noise_level: None | float = None):
        """Return two PyMatching matching graphs for complementary gap calculation.
        
        
        :param noise_level: A probability that represents the noise strength.
            This is passed to ``self.NOISE.get_edge_weights()``.
            If ``None``, all edges have error probability 0 and weight 1.
        
        
        :return matching_1: A PyMatching matching graph with 2 virtual boundary nodes:
            all edges connected to the west (east) virtual boundary node have fault ID 0 (1).
        :return matching_2: Same as ``matching_1`` but the virtual boundary nodes are now detectors,
            numbered in the same order as their fault IDs.
            E.g. if the detector count is 6, then the virtual boundary node corresponding to fault ID 0 is numbered 6,
            and that corresponding to fault ID 1 is numbered 7.
        """
        detector_to_int = self._DETECTOR_TO_INT | {
            v: self.CODE.DETECTOR_COUNT + fault_id
            for v, fault_id in self._WEST_EAST_LOGICALS.items()
        }
        matching_1 = self.get_matching(
            noise_level=noise_level,
            boundary_node_to_fault_id=self._WEST_EAST_LOGICALS,
        )
        matching_2 = self.get_matching(
            noise_level=noise_level,
            detector_to_int=detector_to_int,
        )
        return matching_1, matching_2


"""Module concerning decoder accuracy i.e. failure probabilities.

Available functions:

* monte_carlo
* monte_carlo_results_to_sample_counts
* monte_carlo_pymatching
* monte_carlo_special
* subset_sample
"""

from localuf.sim.accuracy.monte_carlo import monte_carlo, monte_carlo_results_to_sample_counts, \
    monte_carlo_pymatching, monte_carlo_special, subset_sample


import pytest

from localuf.codes import Surface
from localuf.decoders import BUF

@pytest.fixture
def dynamic_buf(sf7F: Surface) -> BUF:
    decoder = BUF(sf7F, dynamic=True)
    decoder.history = []
    return decoder


def test_validate_static(
        buf: BUF,
        validated_static_erasure,
        syndrome7F,
):
    buf.validate(syndrome7F)
    assert len(buf.clusters) == 28
    assert all(bucket == set() for bucket in buf.buckets.values())
    assert buf.erasure == validated_static_erasure


def test_validate_dynamic(
        dynamic_buf: BUF,
        validated_dynamic_erasure,
        syndrome7F,
):
    dynamic_buf.validate(syndrome7F)
    assert len(dynamic_buf.clusters) == 29
    assert all(bucket == set() for bucket in dynamic_buf.buckets.values())
    assert dynamic_buf.erasure == validated_dynamic_erasure


from localuf.codes import Surface
from localuf.decoders.uf import UF


class TestSwimDistance:

    def test_Meister2024_example(self):
        """The example in [arXiv:2405.07433, Figure 2]."""
        decoder = UF(Surface(6, 'code capacity'))
        syndrome = {(2, 1), (3, 3)}
        decoder.decode(syndrome)
        assert decoder.swim_distance() == 1

    def test_syndrome7F(self, uf7F: UF, syndrome7F):
        uf7F.validate(syndrome7F)
        assert uf7F.swim_distance() == 2


class TestWeighCorrection:

    def test_no_noise_level(self, uf7F: UF, syndrome7F):
        uf7F.decode(syndrome7F)
        result = uf7F._weigh_correction()

        assert len(uf7F.correction) == 6
        assert result == len(uf7F.correction)

    def test_a_noise_level(self, uf7F: UF, syndrome7F):
        uf7F.decode(syndrome7F)
        result = uf7F._weigh_correction(noise_level=0.5)

        assert len(uf7F.correction) == 6
        weight = uf7F.CODE.NOISE.log_odds_of_no_flip(0.5)
        assert result == len(uf7F.correction) * weight


import pytest

from localuf.decoders.luf import LUF, MacarNode, _Node, Nodes, Macar, Actis
from localuf.constants import Growth
from localuf.decoders._base_uf import direction

class ConcreteNode(_Node):
    """Instantiable version of `_Node`."""
    def advance(self):
        pass

@pytest.fixture
def ufn(macar: Macar):
    v = (0, 0) if macar.NODES.LUF.CODE.DIMENSION==2 else (0, 0, 0)
    return ConcreteNode(macar.NODES, v)

@pytest.fixture(name="ufn3", params=[
    ("sf3F", "v00"),
    ("sf3T", "v000"),
], ids=['code capacity', 'phenomenological'])
def _ufn3(request):
    """Instance of _Node."""
    sf, v = (request.getfixturevalue(s) for s in request.param)
    decoder = Macar(sf)
    vns = Nodes(decoder)
    return ConcreteNode(vns, v)

@pytest.fixture(name="n3", params=[
    ("sf3F", "v00", Macar),
    ("sf3F", "v00", Actis),
    ("sf3T", "v000", Macar),
    ("sf3T", "v000", Actis),
], ids=[
    'CC-vis',
    'CC-invis',
    'ph-vis',
    'ph-invis',
])
def _n3(request):
    """Instances of MacarNode and ActisNode."""
    sf, v = (request.getfixturevalue(s) for s in request.param[:2])
    _, _, decoder_class = request.param
    decoder = decoder_class(sf)
    ns = decoder.NODES
    return ns.dc[v]

@pytest.fixture
def ufn3F(sf3F, v00):
    decoder = Macar(sf3F)
    vns = Nodes(decoder)
    return ConcreteNode(vns, v00)

@pytest.fixture
def ufn3T(sf3T, v000):
    decoder = Macar(sf3T)
    vns = Nodes(decoder)
    return ConcreteNode(vns, v000)

# @pytest.fixture
# def n3F(luf3F: LUF, v00):
#     nodes = luf3F.nodes
#     return nodes.dc[v00]

# @pytest.fixture
# def n3T(vluf5T: LUF, v000):
#     nodes = luf5T.nodes
#     return nodes.dc[v000]

@pytest.fixture
def n3_with_E(n3: _Node):
    """Node (with correct access) and a pointer string."""
    pointer = 'E'
    e, index = n3.NEIGHBORS[pointer]
    n3.access = {pointer: n3.NODES.dc[e[index]]}
    return n3, pointer

def test_nodes_attribute(ufn3: _Node):
    assert type(ufn3.NODES) is Nodes

def test_v_attribute_2D(ufn3F: _Node, v00):
    assert type(ufn3F.INDEX) is tuple
    assert len(ufn3F.INDEX) == 2
    assert ufn3F.INDEX == v00

def test_v_attribute_3D(ufn3T: _Node, v000):
    assert type(ufn3T.INDEX) is tuple
    assert len(ufn3T.INDEX) == 3
    assert ufn3T.INDEX == v000

def test_id_attribute(ufn: _Node):
    assert type(ufn.ID) is int
    power = ufn.NODES.LUF.CODE.DIMENSION - 1
    assert ufn.ID == 2 * ufn.NODES.LUF.CODE.D**power

def test_neighbors_attribute(ufn: _Node, v00, v000):
    assert type(ufn.NEIGHBORS) is dict
    ph_dc = {
        'W': (((0,-1, 0), v000), 0),
        'E': ((v000, (0, 1, 0)), 1),
        'S': ((v000, (1, 0, 0)), 1),
        'U': ((v000, (0, 0, 1)), 1),
    }
    noise = ufn.NODES.LUF.CODE.NOISE
    if str(noise) == 'code capacity':
        assert ufn.NEIGHBORS == {
            'W': (((0,-1), v00), 0),
            'E': ((v00, (0, 1)), 1),
            'S': ((v00, (1, 0)), 1),
        }
    elif str(noise) == 'phenomenological':
        assert ufn.NEIGHBORS == ph_dc
    elif str(noise) == 'circuit-level':
        assert ufn.NEIGHBORS == ph_dc | {
            'EU': ((v000, (0, 1, 1)), 1),
            'SEU': ((v000, (1, 1, 1)), 1),
        }
    else:
        raise ValueError(f"Unknown noise model {noise}")

@pytest.mark.parametrize("prop", [
    "NODES",
    "INDEX",
    "ID",
    "NEIGHBORS",
])
def test_property_attributes(test_property, ufn3: _Node, prop):
    test_property(ufn3, prop)

def test_reset(ufn3: _Node):
    n3_copy = ConcreteNode(ufn3.NODES, ufn3.INDEX)

    ufn3.defect = True
    ufn3.cid = -1
    ufn3.next_cid = -2
    ufn3.anyon = True
    ufn3.next_anyon = True
    ufn3.pointer = 'N'
    ufn3.active = True
    ufn3.busy = True
    pointer = 'E'
    ufn3.access = {pointer: MacarNode(ufn3.NODES, ufn3.INDEX)}

    ufn3.reset()

    assert not any((
        ufn3.defect,
        ufn3.anyon,
        ufn3.next_anyon,
        ufn3.active,
        ufn3.busy,
    ))
    assert ufn3.cid == ufn3.ID
    assert ufn3.next_cid == ufn3.ID
    assert ufn3.pointer == 'C'
    assert ufn3.access == {}
    
    assert dir(ufn3) == dir(n3_copy)

def test_defect_attribute(ufn3: _Node):
    assert type(ufn3.defect) is bool
    assert ufn3.defect == False

def test_cid_attribute(ufn3: _Node):
    assert type(ufn3.cid) is int
    assert ufn3.cid == ufn3.ID

def test_next_cid_attribute(ufn3: _Node):
    assert type(ufn3.next_cid) is int
    assert ufn3.next_cid == ufn3.ID

def test_anyon_attribute(ufn3: _Node):
    assert type(ufn3.anyon) is bool
    assert ufn3.anyon == False

def test_next_anyon_attribute(ufn3: _Node):
    assert type(ufn3.next_anyon) is bool
    assert ufn3.next_anyon == False
    
def test_pointer_attribute(ufn3: _Node):
    assert type(ufn3.pointer) is str
    assert ufn3.pointer == 'C'
    
def test_active_attribute(ufn3: _Node):
    assert type(ufn3.active) is bool
    assert ufn3.active == False
    
def test_busy_attribute(ufn3: _Node):
    assert type(ufn3.busy) is bool
    assert ufn3.busy == False

def test_access_attribute(ufn3: _Node):
    assert type(ufn3.access) is dict
    assert ufn3.access == {}

def test_make_defect(ufn3: _Node):
    ufn3.make_defect()
    assert all((ufn3.defect, ufn3.anyon, ufn3.active))

def test_growing_inactive(ufn: _Node):
    """Test `growing` when node inactive."""
    ufn.growing()
    assert ufn.busy is False
    assert all(val is Growth.UNGROWN
               for val in ufn.NODES.LUF.growth.values())

def test_growing_active(ufn: _Node):
    """Test `growing` when node active."""
    luf = ufn.NODES.LUF
    ufn.active = True
    for growth_value in [Growth.HALF] + 2*[Growth.FULL]:
        ufn.growing()
        assert ufn.busy is False
        for e in luf.CODE.EDGES:
            if e in luf.CODE.INCIDENT_EDGES[ufn.INDEX]:
                assert luf.growth[e] is growth_value
            else:
                assert luf.growth[e] is Growth.UNGROWN

def test_update_access_2D(macar5F: Macar):
    wce = (0, -1), (0, 0), (0, 1)
    update_access_helper(macar5F, wce)

def test_update_access_3D(macar5T: Macar):
    wce = (0, -1, 0), (0, 0, 0), (0, 1, 0)
    update_access_helper(macar5T, wce)

def update_access_helper(luf: LUF, wce):
    w, c, e = wce
    center = luf.NODES.dc[c]
    luf.growth[w, c] = Growth.FULL
    center.update_access()
    assert center.access == {'W': luf.NODES.dc[w]}
    luf.growth[c, e] = Growth.FULL
    center.update_access()
    assert center.access == {
        'W': luf.NODES.dc[w],
        'E': luf.NODES.dc[e],
    }
    luf.growth[w, c] = Growth.UNGROWN
    center.update_access()
    assert center.access == {'E': luf.NODES.dc[e]}

def test_merging_relay(n3_with_E: tuple[_Node, direction]):
    n3, pointer = n3_with_E
    n3.anyon = True
    n3.pointer = pointer
    n3.merging()
    assert n3.busy
    assert n3.access[pointer].next_anyon
    assert not n3.anyon

def test_merging_never_relay_from_boundary(luf3F: LUF):
    boundary_node = luf3F.NODES.dc[1, -1]
    east_node = luf3F.NODES.dc[1, 0]
    boundary_node.access = {'E': east_node} # type: ignore
    boundary_node.pointer = 'E'
    boundary_node.anyon = True
    east_node.cid = 0
    boundary_node.merging()

    assert not east_node.next_anyon
    assert boundary_node.anyon

    assert boundary_node.busy
    assert boundary_node.pointer == 'E'
    assert boundary_node.next_cid == east_node.cid

def test_merging_no_flood(n3_with_E: tuple[_Node, direction]):
    n3, _ = n3_with_E
    n3.merging()
    assert not n3.busy
    assert n3.pointer == 'C'
    assert n3.next_cid == n3.ID

def test_merging_flood_2D(ns3F: Nodes):
    node = ns3F.dc[1, 0]
    north = ns3F.dc[0, 0]
    west = ns3F.dc[1, -1]
    merging_flood_helper(node, north, west)

def test_merging_flood_3D(ns3T: Nodes):
    node = ns3T.dc[1, 0, 0]
    north = ns3T.dc[0, 0, 0]
    west = ns3T.dc[1, -1, 0]
    merging_flood_helper(node, north, west)

def merging_flood_helper(
        node: _Node,
        north: _Node,
        west: _Node,
):
    node.access = {'N': north, 'W': west}
    node.merging()
    assert node.busy
    assert node.pointer == 'W'
    assert node.next_cid == west.ID

def test_update_after_merge_step(ufn3: _Node):
    ufn3.next_cid = 0
    ufn3.next_anyon = True
    ufn3.update_after_merge_step()
    assert ufn3.cid == 0
    assert ufn3.anyon
    assert not ufn3.next_anyon

def test_presyncing(ufn3: _Node):
    
    ufn3.anyon = True
    ufn3.presyncing()
    assert ufn3.active
    assert not ufn3.busy

    ufn3.anyon = False
    ufn3.presyncing()
    assert not ufn3.active
    assert not ufn3.busy

def test_syncing(n3_with_E: tuple[_Node, direction]):
    vn3, pointer = n3_with_E
    vn3.access[pointer].active = True
    vn3.syncing()
    assert vn3.busy

def test_update_after_sync_step(ufn3: _Node):
    # active, busy = 00
    ufn3.update_after_sync_step()
    assert not ufn3.active
    # active, busy = 01
    ufn3.busy = True
    ufn3.update_after_sync_step()
    assert ufn3.active
    # active, busy = 10
    ufn3.busy = False
    ufn3.update_after_sync_step()
    assert ufn3.active
    # active, busy = 11
    ufn3.busy = True
    ufn3.update_after_sync_step()
    assert ufn3.active


import abc
import collections
from functools import cached_property
from typing import Literal

import networkx as nx

from localuf import constants
from localuf.constants import Growth
from localuf.type_aliases import Edge, Node
from localuf._base_classes import Code
from localuf._schemes import Frugal
from localuf.decoders._base_uf import BaseUF


class UF(BaseUF):
    """The original, algorithmic UF decoder.
    
    Extends ``BaseUF``.
    Incompatible with the frugal decoding scheme.
    
    Additional instance constants:
    
    * ``_FORESTER`` decides whether to maintain a static or dynamic forest.
    * ``_INCLINATION`` a class that decides which boundary the root should be on
        when there is a percolating cluster
        i.e. one that spans opposite boundaries.
        Has the ``update_boundary`` method.
    
    Additional instance attributes:
    
    * ``parents`` maps each node in the decoding graph to its parent node.
    * ``clusters`` maps each root node to the cluster of that root node.
        Initially, EVERY node its own cluster; then, only grow active clusters.
    * ``active_clusters`` the set of clusters
        with odd defect count and no boundary node.
    * ``forest`` the spanning forest after syndrome validation as a list of pairs (edge, node)
        where node is the further endpoint from the root of the tree in the forest.
    * ``digraph`` a NetworkX digraph of ``forest``.
    """

    def __init__(
            self,
            code: Code,
            dynamic: bool = False,
            inclination: Literal['default', 'west'] = 'default',
    ):
        """
        :param code: the code to be decoded.
        :param dynamic: whether the forest is dynamic or static.
            Static forest is when the spanning forest of the erasure
            is grown only after syndrome validation;
            this implementation grows the spanning tree using depth-first search (DFS).
            Dynamic forest is when the spanning forest of the erasure
            is maintained throughout syndrome validation;
            this is equivalent to multi-source breadth-first search (BFS) where each defect is a source.
            Dynamic forests are further explained in https://doi.org/10.13140/RG.2.2.13495.96162, section 5.3.2.
            This ``dynamic`` parameter has a big effect on the complementary gap:
            the gap of dynamic UF closely follows the log odds of success;
            the gap of static UF can be much larger.
            This is because for dynamic UF only, the spanning tree of each cluster
            connects any two defects in a low-weight path.
        :param inclination: decides which boundary the root should be on when there is a percolating cluster.
            'default' means prefer neither east nor west, essentially random.
            'west' defines the root of a percolating cluster always at the west boundary.
        """
        if isinstance(code.SCHEME, Frugal):
            raise ValueError('UF incompatible with frugal scheme.')
        self.history: list[UF]
        self.correction = set()
        self._FORESTER = _DynamicForester(self) if dynamic else _StaticForester(self)
        inclination_class = _DefaultInclination if inclination == 'default' else _WestInclination
        self._INCLINATION = inclination_class(code.LONG_AXIS)
        super().__init__(code)
        self.reset()

    def __repr__(self):
        return f'UF(code={self.CODE}, syndrome={self.syndrome})'
    
    def reset(self, no_boundaries=False):
        """Factory reset.
        
        
        :param no_boundaries: whether to treat all boundary nodes as detectors too.
        """
        super().reset()
        self.syndrome = set()
        # note: following line is faster than dict(zip(nodes, nodes))
        self.parents = {v: v for v in self.CODE.NODES}
        self.clusters = {v: _Cluster(self, v, no_boundaries=no_boundaries) for v in self.CODE.NODES}
        self.active_clusters: set[_Cluster] = set()
        try: del self.changed_edges
        except AttributeError: pass
        try: del self.forest
        except AttributeError: pass
        try: del self.digraph
        except AttributeError: pass

    def load(self, syndrome: set[Node]):
        """Load a new syndrome."""
        self.syndrome = syndrome
        for v in syndrome:
            cluster = self.clusters[v]
            cluster.odd = True
            self.active_clusters.add(cluster)
    
    def decode(
            self,
            syndrome: set[Node],
            draw=False,
            fig_width: float | None = None,
    ):
        """Additional inputs over ``Decoder.decode()``:
        
        * ``fig_width`` figure width.
        """
        self.validate(syndrome, log_history=draw)
        self.peel()
        if draw:
            self.draw_decode(fig_width=fig_width)

    def complementary_gap(
            self,
            noise_level: None | float = None,
            draw=False,
            fig_width: float | None = None,
            fig_height: float | None = None,
            **kwargs_for_networkx_draw,
    ):
        """Calculate the complementary gap after decoding.
        
        Assumes ``self.decode()`` has already been called.
        If ``self.CODE.MERGED_EQUIVALENT_BOUNDARY_NODES`` is ``False``,
        raises a ``ValueError``.
        
        
        :param noise_level: a probability representing the noise strength.
            This is needed to define nonuniform edge weights of the decoding graph
            in the circuit-level noise model.
            If ``None``, all edges are assumed to have weight 1.
        :param draw: whether to draw the original and complementary corrections.
        :param fig_width: figure width.
        :param fig_height: figure height.
        :param kwargs_for_networkx_draw: passed to ``networkx.draw()``.
        
        
        :returns: ``weight_2 - weight_1`` the complementary gap.
        
        Side effects:
        
        * ``self`` ends up in the final state after computing the complementary correction.
        * All clusters are blind to boundary nodes
            so will stop growing only once they have an odd defect count.
        """
        if not self.CODE.MERGED_EQUIVALENT_BOUNDARY_NODES:
            raise ValueError('Complementary gap requires all equivalent boundary nodes in the decoding graph be merged.')
        if draw:
            from matplotlib import pyplot as plt
            if fig_width is None: fig_width = self._FIG_WIDTH
            if fig_height is None: fig_height = self._FIG_HEIGHT
            plt.figure(figsize=(2*fig_width, fig_height))
            plt.subplot(121)
            self.draw_peel(**kwargs_for_networkx_draw)
        weight_1 = self._weigh_correction(noise_level=noise_level)
        complementary_syndrome = self.CODE.get_verbose_syndrome(
            self.correction).symmetric_difference(self.CODE.BOUNDARY_NODES)
        self.reset(no_boundaries=True)
        self.decode(complementary_syndrome)
        if draw:
            plt.subplot(122)
            self.draw_peel(**kwargs_for_networkx_draw)
            plt.tight_layout()
        weight_2 = self._weigh_correction(noise_level=noise_level)
        return weight_2 - weight_1
    
    def _weigh_correction(self, noise_level: None | float = None) -> float:
        """Weigh ``self.correction``.
        
        
        :param noise_level: A probability representing the noise strength.
            This is needed to define nonuniform edge weights of the decoding graph
            in the circuit-level noise model.
            If not specified, all edges are assumed to have weight 1.
        
        
        :return weight: The sum of the weights of the edges in ``self.correction``.
        """
        edge_weights = self.CODE.NOISE.get_edge_weights(noise_level)
        return sum(edge_weights[e][1] for e in self.correction)
    
    def unclustered_node_fraction(self):
        """Compute the Unclustered Node Fraction DCS.
        
        
        :return fraction: The fraction of nodes that are not in a cluster.
        """
        clustered_node_count = sum(cluster.size
            for cluster in self.clusters.values() if cluster.size > 1)
        return 1 - clustered_node_count/self.CODE.NODE_COUNT

    # SYNDROME VALIDATION

    def validate(
            self,
            syndrome: set[Node],
            log_history=False,
    ):
        """Grow all active clusters until they are inactive."""
        self.load(syndrome)
        if log_history: self.init_history()
        while self.active_clusters:
            self._growth_round(log_history)

    def _growth_round(
            self,
            log_history: bool,
            clusters_to_grow: 'set[_Cluster] | None' = None,
    ):
        """Grow each active cluster once."""
        if clusters_to_grow is None:
            clusters_to_grow = self.active_clusters
        merge_ls: list[Edge] = []
        changed_edges: set[Edge] = set()
        for cluster in clusters_to_grow:
            self._grow(cluster, merge_ls, changed_edges)
        for (u, v) in merge_ls:
            cu = self.clusters[self._find(u)]
            cv = self.clusters[self._find(v)]
            # remove edge (u, v) from all visions
            cu.vision.remove((u, v))
            cv.vision.discard((u, v))  # discard instead of remove as cu may be cv
            self._FORESTER.merge(cu, cv, (u, v))
        if log_history:
            self.changed_edges: set[Edge] = changed_edges
            self.append_history()

    def _grow(
            self,
            cluster: '_Cluster',
            merge_ls: list[Edge],
            changed_edges: set[Edge],
    ):
        """Grow a cluster once."""
        changed_edges.update(cluster.vision)
        for e in cluster.vision:
            self.growth[e] += Growth.INCREMENT
            if self.growth[e] is Growth.FULL:
                merge_ls.append(e)

    def _find(self, v: Node):
        """Find root of ``v`` (w/ path compression).
        
        From https://www.geeksforgeeks.org/union-by-rank-and-path-compression-in-union-find-algorithm/.
        """
        if self.parents[v] != v:
            self.parents[v] = self._find(self.parents[v])
        return self.parents[v]
    
    def static_merge(self, cu: '_Cluster', cv: '_Cluster'):
        """Merge clusters ``cu`` and ``cv`` along edge ``uv`` for static UF."""
        if cu != cv:
            self._union(cu, cv)

    def dynamic_merge(self, cu: '_Cluster', cv: '_Cluster', e: Edge):
        """Merge clusters ``cu`` and ``cv`` along edge ``uv`` for dynamic UF."""
        if cu == cv or (cu.boundary and cv.boundary):
            self.growth[e] = Growth.BURNT
        else:
            self._union(cu, cv)

    def _union(self, cu: '_Cluster', cv: '_Cluster'):
        """Union clusters ``cu`` with ``cv`` (by weight)."""
        if cu.size >= cv.size:
            larger, smaller = cu, cv
        else:
            larger, smaller = cv, cu
        self.parents[smaller.root] = larger.root
        larger.size += smaller.size
        larger.odd ^= smaller.odd  # logical XOR
        old_larger_vision_length = len(larger.vision) + 1  # +1 as we already removed (u, v) from all visions
        larger.vision.update(smaller.vision)
        self._INCLINATION.update_boundary(larger, smaller)
        self._update_self_after_union(larger, smaller, old_larger_vision_length)

    def _update_self_after_union(
            self,
            larger: '_Cluster',
            smaller: '_Cluster',
            old_larger_vision_length: int,
    ):
        """Update attributes ``clusters`` and ``active_clusters`` after union of larger w/ smaller."""

        # DELETE SMALLER CLUSTER
        del self.clusters[smaller.root]
        # discard instead of remove as smaller may be a node not in syndrome
        self.active_clusters.discard(smaller)

        # UPDATE `active_clusters` SET
        if larger.odd and not larger.boundary:
            # need this line as an even cluster could become odd again by union w/ smaller odd cluster
            self.active_clusters.add(larger)
        else:
            # discard instead of remove as larger could have had boundary before union, hence was not active
            self.active_clusters.discard(larger)

    # PEELING DECODER
    
    def peel(self):
        """Peel the validated erasure."""
        self.forest = self._make_forest()
        modified_syndrome = self.syndrome.copy()
        for e, u in reversed(self.forest):
            if u in modified_syndrome:
                self.correction.add(e)
                modified_syndrome.symmetric_difference_update(e)
    
    def _make_forest(self):
        """Return a spanning forest of the validated erasure.
        
        
        :returns: ``forest`` a list of pairs (edge, node) where the node is the further endpoint from the tree root.
        """
        forest: list[tuple[Edge, Node]] = []
        for cluster in self.clusters.values():
            tree_root = cluster.boundary if cluster.boundary else cluster.root
            forest += self._make_tree(tree_root)
        return forest

    def _make_tree(self, tree_root: Node):
        """Return a spanning tree of the validated cluster, from ``tree_root``.
        
        
        :returns: ``tree`` a list of pairs (edge, node) where the node is the further endpoint from the root.
        """
        discovered = {tree_root}
        stack = collections.deque([tree_root])
        tree: list[tuple[Edge, Node]] = []
        while stack:  # each edge is considered twice due to its two endpoints
            u = stack.pop()
            for e in self.erasure.intersection(self.CODE.INCIDENT_EDGES[u]):
                v = self.CODE.traverse_edge(e, u)
                if v not in discovered:
                # could also check nb not surface boundary (when not dynamic), but not needed
                    discovered.add(v)
                    stack.append(v)
                    tree.append((e, v))
        return tree

    # DRAWERS

    def draw_growth(self, **kwargs):
        return super().draw_growth(**kwargs, syndrome=self.syndrome)

    @cached_property
    def digraph(self):
        """Return a NetworkX digraph representing the spanning forest after syndrome validation.
        
        Note: Requires calling ``peel()`` first.
        """
        dig = nx.DiGraph()
        dig.add_nodes_from(self.CODE.NODES)
        for e, u in self.forest:
            v = self.CODE.traverse_edge(e, u)
            dig.add_edge(u, v)
        return dig

    def draw_forest(
        self,
        x_offset=constants.DEFAULT_X_OFFSET,
        node_size=constants.SMALL,
        width=constants.WIDE_MEDIUM,
        edge_color = 'g',
        boundary_color=constants.BLUE,
        defect_color=constants.RED,
        nondefect_color=constants.GREEN,
        show_boundary_defects=True,
        **kwargs_for_networkx_draw,
    ):
        """Draw spanning forest.
        
        ``kwargs_for_networkx_draw`` passed to ``networkx.draw()``.
        
        Note: Not as informative as draw_peel().
        """
        dig = self.digraph.reverse()
        pos = self.CODE.get_pos(x_offset)
        node_color = self.CODE.get_node_color(
            self.syndrome,
            boundary_color=boundary_color,
            defect_color=defect_color,
            nondefect_color=nondefect_color,
            show_boundary_defects=show_boundary_defects,
        )
        return nx.draw(
            dig,
            pos=pos,
            node_size=node_size,
            node_color=node_color,
            width=width,
            edge_color=edge_color,
            **kwargs_for_networkx_draw,
        )

    def draw_peel(
        self,
        x_offset=constants.DEFAULT_X_OFFSET,
        node_size=constants.SMALL,
        width=constants.WIDE_MEDIUM,
        correction_color='k',
        boundary_color=constants.BLUE,
        defect_color=constants.RED,
        nondefect_color=constants.GREEN,
        show_boundary_defects=True,
        **kwargs_for_networkx_draw,
    ):
        """Draw forest and correction from peeling.
        
        ``kwargs_for_networkx_draw`` passed to ``networkx.draw()``.
        
        Note: Requires calling ``peel()`` first.
        """
        dig = self.digraph
        pos = self.CODE.get_pos(x_offset)
        node_color = self.CODE.get_node_color(
            self.syndrome,
            boundary_color=boundary_color,
            defect_color=defect_color,
            nondefect_color=nondefect_color,
            show_boundary_defects=show_boundary_defects,
        )
        edge_color = [
            correction_color if ((u, v) in self.correction) or ((v, u) in self.correction) else
            constants.GRAY for u, v in dig.edges
        ]
        return nx.draw(
            dig,
            pos=pos,
            with_labels=False,
            node_size=node_size,
            node_color=node_color,
            width=width,
            edge_color=edge_color,
            **kwargs_for_networkx_draw,
        )

    def draw_decode(
            self,
            fig_width: float | None = None,
            fig_height: float | None = None,
            **kwargs_for_networkx_draw,
    ):
        from matplotlib import pyplot as plt
        if fig_width is None: fig_width = self._FIG_WIDTH
        if fig_height is None: fig_height = self._FIG_HEIGHT
        n_plots = len(self.history) + 1
        plt.figure(figsize=(fig_width, fig_height*n_plots))
        for k, older_self in enumerate(self.history, start=1):
            plt.subplot(n_plots, 1, k)
            older_self.draw_growth(
                highlighted_edges=older_self.changed_edges,
                **kwargs_for_networkx_draw,
            )
        plt.subplot(n_plots, 1, n_plots)
        self.draw_peel(**kwargs_for_networkx_draw)
        plt.tight_layout()

    @property
    def _FIG_WIDTH(self):
        return max(1, self.CODE.D * (2*self.CODE.DIMENSION - 3) / 3)
    
    @property
    def _FIG_HEIGHT(self):
        d = self.CODE.D
        h = self.CODE.SCHEME.WINDOW_HEIGHT
        w = self._FIG_WIDTH
        if (n:=self.CODE.DIMENSION) == 1:
            return w / d
        elif n == 2:
            return w * max(d, h) / d
        else:
            return w * h / d


class BaseCluster(abc.ABC):
    """Abstract base class for the cluster in UF.
    
    Mathematically, a cluster is a connected subgraph of the decoding graph.
    
    Attributes:
    
    * ``root`` the representative of the cluster.
    * ``size`` the number of nodes in the cluster.
    * ``odd`` a Boolean to say if the number of defects in the cluster is odd.
    * ``boundary`` the node in the cluster that is the surface boundary if it exists; else, None.
        If ``cluster.boundary`` not None, cluster will never be active in future.
        If syndrome validation not dynamic, boundary may not be unique!
    """

    def __init__(self, uf: UF, root: Node, no_boundaries=False):
        self.root = root
        self.size = 1
        self.odd = False
        self.boundary = root if (not no_boundaries and uf.CODE.is_boundary(root)) else None

    def __str__(self) -> str:
        return f'cluster(root={self.root}, size={self.size}, odd={self.odd})'


class _Cluster(BaseCluster):
    """A UF cluster that tracks which edges to grow next.
    
    Extends ``BaseCluster``.
    
    Additional attributes:
    
    * ``vision`` a set of active edges incident to the cluster
        i.e. edges to be grown by the cluster in next growth round.
    """

    def __init__(self, uf: UF, root: Node, no_boundaries=False):
        super().__init__(uf, root, no_boundaries=no_boundaries)
        self.vision = uf.CODE.INCIDENT_EDGES[root].copy()
    

class _Inclination(abc.ABC):
    """Decides which boundary the root should be on when there is a percolating cluster."""

    def __init__(self, long_axis: int):
        self._LONG_AXIS = long_axis

    @abc.abstractmethod
    def update_boundary(self, larger: BaseCluster, smaller: BaseCluster):
        """Update the ``boundary`` attribute of the cluster ``larger``."""


class _DefaultInclination(_Inclination):
    """Prefer neither east nor west."""

    def update_boundary(self, larger: BaseCluster, smaller: BaseCluster):
        if (not larger.boundary) and smaller.boundary:
            larger.boundary = smaller.boundary


class _WestInclination(_Inclination):
    """Define the root of a percolating cluster at the west boundary."""

    def update_boundary(self, larger: BaseCluster, smaller: BaseCluster):
        if smaller.boundary and (
            (not larger.boundary) or
            (smaller.boundary[self._LONG_AXIS] < larger.boundary[self._LONG_AXIS])
        ):
            larger.boundary = smaller.boundary


class _Forester(abc.ABC):
    """Decides whether to maintain a static or dynamic forest.
    
    See the ``dynamic`` kwarg of ``UF.__init__`` for details.
    """

    def __init__(self, uf: UF):
        self.UF = uf

    @abc.abstractmethod
    def merge(self, cu: _Cluster, cv: _Cluster, e: Edge):
        """Merge clusters ``cu`` and ``cv`` along edge ``e``."""


class _StaticForester(_Forester):
    """Maintains a static forest.
    
    Extends ``_Forester``.
    """

    def merge(self, cu, cv, e):
        self.UF.static_merge(cu, cv)


class _DynamicForester(_Forester):
    """Maintains a dynamic forest.
    
    Extends ``_Forester``.
    """

    def merge(self, cu, cv, e):
        self.UF.dynamic_merge(cu, cv, e)


import math
from unittest import mock

import pytest

from localuf import Repetition, Surface
from localuf.type_aliases import Edge
from localuf.decoders import Snowflake
from localuf._schemes import Frugal
from localuf._pairs import LogicalCounter


@pytest.fixture(params=range(3, 11, 2), ids=lambda x: f"d{x}")
def rp_frugal(request):
    rp = Repetition(
        request.param,
        noise='phenomenological',
        scheme='frugal',
    )
    return rp.SCHEME

@pytest.fixture(params=range(3, 11, 2), ids=lambda x: f"d{x}")
def sf_frugal(request):
    sf = Surface(
        request.param,
        noise='phenomenological',
        scheme='frugal',
    )
    return sf.SCHEME

@pytest.fixture
def frugal3(rp3_frugal: Repetition):
    return rp3_frugal.SCHEME


def test_LOGICAL_COUNTER_attribute(frugal3: Frugal):
    assert type(frugal3._LOGICAL_COUNTER) is LogicalCounter


def test_COMMIT_HEIGHT_attribute(rp_frugal: Frugal):
    assert rp_frugal._COMMIT_HEIGHT == 1


def test_BUFFER_HEIGHT_attribute(rp_frugal: Frugal):
    assert rp_frugal._BUFFER_HEIGHT == 2*(rp_frugal._CODE.D//2)


def test_reset(frugal3: Frugal):
    e = ((-1, 0), (0, 0))
    frugal3.error = {e}
    frugal3.step_counts = [2]
    frugal3._future_boundary_syndrome = {(0, 3)}
    with mock.patch("localuf._pairs.Pairs.reset") as mock_reset:
        frugal3.reset()
        mock_reset.assert_called_once_with()
    assert frugal3.error == set()
    assert frugal3.step_counts == []
    assert frugal3._future_boundary_syndrome == set()


def test_get_logical_error_repetition(rp_frugal: Frugal):

    assert not rp_frugal.get_logical_error()
    assert rp_frugal.pairs._dc == {}
    
    u, v = (-1, 0), (0, 0)
    rp_frugal.pairs._dc = {u: v, v: u}
    assert not rp_frugal.get_logical_error()
    uu, vv = (-1, -1), (0, -1)
    assert rp_frugal.pairs._dc == {uu: vv, vv: uu}

    u, v = (-1, 0), (rp_frugal._CODE.D-1, 0)
    rp_frugal.pairs._dc = {u: v, v: u}
    assert rp_frugal.get_logical_error()
    assert rp_frugal.pairs._dc == {}


def test_get_logical_error_surface(sf_frugal: Frugal):
    
    assert not sf_frugal.get_logical_error()
    assert sf_frugal.pairs._dc == {}

    u, v = (0, -1, 0), (0, 0, 0)
    sf_frugal.pairs._dc = {u: v, v: u}
    assert not sf_frugal.get_logical_error()
    uu, vv = (0, -1, -1), (0, 0, -1)
    assert sf_frugal.pairs._dc == {uu: vv, vv: uu}

    u, v = (0, -1, 0), (0, sf_frugal._CODE.D-1, 0)
    sf_frugal.pairs._dc = {u: v, v: u}
    assert sf_frugal.get_logical_error()
    assert sf_frugal.pairs._dc == {}

    # unique to surface: north--south edge
    u, v = (0, 0, 0), (1, 0, 0)
    sf_frugal.pairs._dc = {u: v, v: u}
    assert not sf_frugal.get_logical_error()
    uu, vv = (0, 0, -1), (1, 0, -1)
    assert sf_frugal.pairs._dc == {uu: vv, vv: uu}


def test_advance(frugal3: Frugal):
    noise_level = 1/2
    with (
        mock.patch("localuf._schemes.Frugal._raise_window") as rw,
        mock.patch(
            f"localuf.codes.Repetition.make_error",
            return_value=set(),
        ) as me,
        mock.patch(
            f"localuf._schemes.Frugal._load",
            return_value=set(),
        ) as mock_load,
        mock.patch(
            f"localuf.decoders.snowflake.Snowflake.decode",
            return_value=2,
        ) as mock_decode,
    ):
        assert frugal3.advance(noise_level, Snowflake(frugal3._CODE)) == 2
        rw.assert_called_once_with()
        me.assert_called_once_with(noise_level, exclude_future_boundary=False)
        mock_load.assert_called_once_with(set())
        mock_decode.assert_called_once_with(set())


def test_raise_window(rp_frugal: Frugal):

    raise_window_helper(rp_frugal)
    
    e, f, g = (
        ((-1, 0), (0, 0)),
        ((-1, 1), (0, 1)),
        ((0, 0), (0, 1)),
    )
    rp_frugal.error = {e, f, g}
    rp_frugal._raise_window()
    assert rp_frugal.error == {e}
    assert rp_frugal.pairs._dc == {
        (-1, 0): (0, 1),
        (0, 1): (-1, 0),
    }
    rp_frugal.pairs._dc = {
        (-1, -1): (0, 0),
        (0, 0): (-1, -1),
    }
    rp_frugal._raise_window()
    assert rp_frugal.error == set()
    assert rp_frugal.pairs._dc == {(-1, 0): (-1, -1), (-1, -1): (-1, 0)}

def raise_window_helper(frugal: Frugal):
    frugal._raise_window()
    assert frugal.error == set()
    assert frugal.pairs._dc == {}


def test_load(rp_frugal: Frugal):
    h = rp_frugal.WINDOW_HEIGHT
    edges = (
        ((-1, h-2), (0, h-2)),
        ((-1, h-1), (0, h-1)),
        ((0, h-1), (1, h-1)),
        ((1, h-1), (2, h-1)),
        ((0, h-1), (0, h)),
    )
    error: set[Edge] = set(edges[1:])
    rp_frugal.error = set(edges[:1])
    rp_frugal._future_boundary_syndrome = {(1, h)}
    syndrome = rp_frugal._load(error)
    assert rp_frugal.error == set(edges)
    assert syndrome == {
        (0, h-1),
        (1, h-1),
    } if rp_frugal._CODE.D == 3 else {
        (0, h-1),
        (1, h-1),
        (2, h-1),
    }
    assert rp_frugal._future_boundary_syndrome == {(0, h)}


class TestRun:

    def test_zero_slenderness(self, frugal3: Frugal):
        decoder = Snowflake(frugal3._CODE)
        with pytest.raises(ValueError):
            frugal3.run(decoder, 1, 0)

    @pytest.mark.parametrize("n", range(1, 4))
    def test_run(self, frugal3: Frugal, n: int):
        d = frugal3._CODE.D
        decoder = Snowflake(frugal3._CODE)
        noise_level = 1/2
        transient_count = math.ceil(frugal3.WINDOW_HEIGHT / frugal3._COMMIT_HEIGHT)
        steady_state_raise_count = (n+1) * d
        cleanse_count = 2 * frugal3.WINDOW_HEIGHT
        raise_count = transient_count + steady_state_raise_count + cleanse_count
        with (
            mock.patch("localuf._schemes.Frugal.reset") as mock_reset,
            mock.patch("localuf.decoders.Snowflake.reset") as snow_reset,
            mock.patch("localuf._schemes.Frugal.advance") as mock_advance,
            mock.patch(
                "localuf._schemes.Frugal.get_logical_error",
                return_value=1,
            ) as gle,
            mock.patch("localuf.decoders.snowflake.Snowflake.init_history") as ih,
            mock.patch("localuf.decoders.snowflake.Snowflake.draw_decode") as dd,
        ):
            failure_count, slenderness = frugal3.run(decoder, noise_level, n)
            assert failure_count == raise_count
            assert slenderness == (transient_count + steady_state_raise_count) / d
            mock_reset.assert_called_once_with()
            snow_reset.assert_called_once_with()
            assert mock_advance.call_args_list == \
                (transient_count + d*n + d-1) * [mock.call(
                    noise_level,
                    decoder,
                    exclude_future_boundary=False,
                    log_history=False,
                    confidence_scores=(),
                    noise_level_for_priors=noise_level,
                    time_only='merging',
                )] \
                + [mock.call(
                    noise_level,
                    decoder,
                    exclude_future_boundary=True,
                    log_history=False,
                    confidence_scores=(),
                    noise_level_for_priors=noise_level,
                    time_only='merging',
                )] \
                + cleanse_count * [mock.call(
                    0,
                    decoder,
                    exclude_future_boundary=False,
                    log_history=False,
                    confidence_scores=(),
                    noise_level_for_priors=noise_level,
                    time_only='merging',
                )]
            assert gle.call_args_list == raise_count * [mock.call()]
            ih.assert_not_called()
            dd.assert_not_called()

            frugal3.run(decoder, noise_level, n, draw='fine')
            ih.assert_called_once_with()
            dd.assert_called_once_with()


"""Module containing type aliases."""

from typing import Literal
from collections.abc import Sequence

import numpy as np
import numpy.typing as npt


Node = tuple[int, ...]
"""Node index."""
Edge = tuple[Node, Node]
"""Edge index."""
Coord = tuple
"""Positional coordinate."""
DecodingScheme = Literal[
    'batch',
    'global batch',
    'forward',
    'frugal',
]
"""Decoding scheme."""
NoiseModel = Literal[
    'code capacity',
    'phenomenological',
    'circuit-level',
]
StreamingNoiseModel = Literal[
    'phenomenological',
    'circuit-level',
]
"""Noise model."""
Parametrization = Literal[
    'standard',
    'balanced',
    'ion trap',
]
"""Affects only circuit-level noise model."""
EdgeType = Literal[
    'S',
    'E westmost',
    'E bulk',
    'E eastmost',
    'U 3',
    'U 4',
    'SD',
    'EU west corners',
    'EU east corners',
    'EU edge',
    'EU centre',
    'SEU',
]
"""Edge type in circuit-level decoding graph for surface code."""
FourFloats = tuple[
    float,
    float,
    float,
    float,
]
"""Tuple of 4 floats."""
FourInts = tuple[
    int,
    int,
    int,
    int,
]
"""Tuple of 4 integers."""
MultiplicityVector = npt.NDArray[np.int_]
"""NumPy array of 4 integers."""
FloatSequence = Sequence[float] | npt.NDArray[np.float64]
IntSequence = Sequence[int] | npt.NDArray[np.int_]


from localuf.constants import Growth

def test_iadd():
    """Test the __iadd__ method of Growth enum."""
    assert Growth.BURNT + Growth.INCREMENT == Growth.UNGROWN
    assert Growth.UNGROWN + Growth.INCREMENT == Growth.HALF
    assert Growth.HALF + Growth.INCREMENT == Growth.FULL

def test_as_float():
    """Test the as_float property of Growth enum."""
    burnt = Growth.BURNT
    assert burnt.as_float == 1.0
    ungrown = Growth.UNGROWN
    assert ungrown.as_float == 0.0
    half = Growth.HALF
    assert half.as_float == 0.5
    full = Growth.FULL
    assert full.as_float == 1.0


import itertools

import pytest
import networkx as nx

from localuf import Surface
from localuf.type_aliases import Node
from localuf.decoders.luf import Macar, Actis
from localuf.constants import Growth
from localuf.decoders.luf.constants import Stage

@pytest.mark.parametrize("macar, syndrome", [
    ("macar5F", "syndrome5F"),
    ("macar5T", "syndrome5T"),
])
def test_not_any_busy_after_growing(macar: Macar, syndrome: set[Node], request):
    macar = request.getfixturevalue(macar)
    syndrome = request.getfixturevalue(syndrome)
    macar.NODES.load(syndrome)
    macar.NODES.advance()
    assert not any(node.busy for node in macar.NODES.dc.values())

@pytest.mark.parametrize("macar, syndrome", [
    ("macar5F", "syndrome5F"),
    ("macar5T", "syndrome5T"),
])
def test_not_any_busy_after_presyncing(macar: Macar, syndrome: set[Node], request):
    macar = request.getfixturevalue(macar)
    syndrome = request.getfixturevalue(syndrome)
    macar.NODES.load(syndrome)
    while not macar.NODES.valid:
        check = macar.CONTROLLER.stage is Stage.PRESYNCING
        macar._advance()
        if check:
            assert not any(node.busy for node in macar.NODES.dc.values())
            assert macar.CONTROLLER.stage is Stage.SYNCING

def test_presyncing_macar(macar: Macar):
    for _ in itertools.repeat(None, 2):
        macar._advance()
    # controller stage now PRESYNC
    for node in macar.NODES.dc.values():
        node.active = True
    macar._advance()
    assert all(not node.active for node in macar.NODES.dc.values())

def test_presyncing_actis(actis: Actis):
    actis.CONTROLLER.stage = Stage.PRESYNCING
    for node in actis.NODES.dc.values():
        node.stage = Stage.PRESYNCING # type: ignore
        node.active = True
    actis._advance()
    assert all(not node.active for node in actis.NODES.dc.values())

def test_validate_single_isolated_error(macar3F: Macar, actis3Fu: Actis):
    syndrome = {(0, 0), (0, 1)}
    erasure = {((0, 0), (0, 1))}

    n_steps = macar3F.validate(syndrome)
    assert n_steps == 6
    assert macar3F.erasure == erasure

    n_steps = actis3Fu.validate(syndrome)
    assert n_steps == 28
    assert actis3Fu.erasure == erasure

def test_validate_vdash_shape(macar3F: Macar, actis3Fu: Actis):
    syndrome = {
        (0, 0),
        (1, 0),
        (1, 1),
        (2, 0),
    }
    erasure = {
        ((0, 0), (1, 0)),
        ((1, 0), (1, 1)),
        ((1, 0), (2, 0)),
    }

    n_steps = macar3F.validate(syndrome)
    assert n_steps == 6
    assert macar3F.erasure == erasure

    n_steps = actis3Fu.validate(syndrome)
    assert n_steps == 29
    assert actis3Fu.erasure == erasure

def test_validate_sixpack(
        sf3F: Surface,
        macar3F: Macar,
        actis3Fu: Actis,
):
    syndrome = {v for v in sf3F.NODES if not sf3F.is_boundary(v)}
    erasure = {e for e in sf3F.EDGES if not any(
        sf3F.is_boundary(v) for v in e
    )}

    n_steps = macar3F.validate(syndrome)
    assert n_steps == 8
    assert macar3F.erasure == erasure

    n_steps = actis3Fu.validate(syndrome)
    assert n_steps == 31
    assert actis3Fu.erasure == erasure

def test_validate_perp_shape(macar5F: Macar, actis5Fu: Actis):
    syndrome = {
        (0, 1),
        (1, 0),
        (1, 1),
        (1, 2),
    }
    erasure = {
        ((0, 1), (1, 1)),
        ((1, 0), (1, 1)),
        ((1, 1), (1, 2)),
    }

    n_steps = macar5F.validate(syndrome)
    assert n_steps == 8
    assert macar5F.erasure == erasure

    n_steps = actis5Fu.validate(syndrome)
    assert n_steps == 43
    assert actis5Fu.erasure == erasure

def test_validate_diagonal(sf5F: Surface):
    macar = Macar(sf5F)
    syndrome = {
        (1, 1),
        (2, 2),
    }
    n_steps = macar.validate(syndrome)
    assert n_steps == 12
    assert macar.erasure == {
        ((0, 1), (1, 1)),
        ((1, 0), (1, 1)),
        ((1, 1), (1, 2)),
        ((1, 1), (2, 1)),
        ((1, 2), (2, 2)),
        ((2, 1), (2, 2)),
        ((2, 2), (2, 3)),
        ((2, 2), (3, 2)),
    }

def test_validate_J_shape(sf5F: Surface):
    macar = Macar(sf5F)
    syndrome = {
        (0, 1),
        (0, 2),
        (1, 2),
        (2, 2),
        (3, 0),
        (3, 1),
        (3, 2),
    }
    n_steps = macar.validate(syndrome)
    assert n_steps == 36
    assert macar.erasure == {
        ((0, 0), (0, 1)),
        ((0, 1), (0, 2)),
        ((0, 1), (1, 1)),
        ((0, 2), (0, 3)),
        ((0, 2), (1, 2)),
        ((1, 1), (1, 2)),
        ((1, 2), (1, 3)),
        ((1, 2), (2, 2)),
        ((2, 0), (3, 0)),
        ((2, 1), (2, 2)),
        ((2, 1), (3, 1)),
        ((2, 2), (2, 3)),
        ((2, 2), (3, 2)),
        ((3, -1), (3, 0)),
        ((3, 0), (3, 1)),
        ((3, 0), (4, 0)),
        ((3, 1), (3, 2)),
        ((3, 1), (4, 1)),
        ((3, 2), (3, 3)),
        ((3, 2), (4, 2))
    }

def test_cid_and_active_causality(sf5F: Surface):
    """Ensure causality i.e. info travels at 1 edge per timestep.
    
    Specifically, info is `cid` & `active`.
    This is ensured via `next_cid` & `busy`, respectively.
    """
    macar = Macar(sf5F)
    syndrome_3_row = {
        (0, 0),
        (0, 1),
        (0, 2),
    }
    n_steps = macar.validate(syndrome_3_row)
    assert n_steps == 17
    assert macar.erasure == {
        ((0, -1), (0, 0)),
        ((0, 0), (0, 1)),
        ((0, 0), (1, 0)),
        ((0, 1), (0, 2)),
        ((0, 1), (1, 1)),
        ((0, 2), (0, 3)),
        ((0, 2), (1, 2))
    }

def test_anyon_causality(sf5F: Surface):
    """Ensure causality of anyon info.
    
    Ensured via `next_anyon`.
    Cluster touches both boundaries."""
    macar = Macar(sf5F)
    syndrome_rightharpoonup = {
        (0, 3),
        (1, 0),
        (1, 1),
        (1, 2),
        (1, 3),
    }
    n_steps = macar.validate(syndrome_rightharpoonup)
    assert n_steps == 26
    assert macar.erasure == {
        ((0, 0), (1, 0)),
        ((0, 1), (1, 1)),
        ((0, 2), (0, 3)),
        ((0, 2), (1, 2)),
        ((0, 3), (0, 4)),
        ((0, 3), (1, 3)),
        ((1, -1), (1, 0)),
        ((1, 0), (1, 1)),
        ((1, 0), (2, 0)),
        ((1, 1), (1, 2)),
        ((1, 1), (2, 1)),
        ((1, 2), (1, 3)),
        ((1, 2), (2, 2)),
        ((1, 3), (1, 4)),
        ((1, 3), (2, 3))
    }

def test_busy_between_syncing(sf5F: Surface, syndrome5F: set[Node]):
    """Ensure `busy` from one syncing stage affects not `active` for next.

    If `busy` remains `True` from previous syncing stage to next,
    upper cluster grows even after touching boundary.
    Both `n_steps` and `decoder.erasure` will then be incorrect.
    """
    decoder = Macar(sf5F)
    n_steps = decoder.validate(syndrome5F)
    assert n_steps == 37
    assert decoder.erasure == {
        ((0, -1), (0, 0)),
        ((0, 0), (0, 1)),
        ((0, 1), (0, 2)),
        ((0, 2), (0, 3)),

        ((0, 0), (1, 0)),
        ((0, 1), (1, 1)),
        ((0, 2), (1, 2)),


        ((2, 1), (3, 1)),

        ((3, 0), (3, 1)),
        ((3, 1), (3, 2)),

        ((3, 0), (4, 0)),
        ((3, 1), (4, 1)),
        ((3, 2), (4, 2)),

        ((4, -1), (4, 0)),
        ((4, 0), (4, 1)),
        ((4, 1), (4, 2)),
        ((4, 2), (4, 3)),
    }

def test_validate_big_P_shape(sf7F: Surface):
    decoder = Macar(sf7F)
    syndrome = {
    (0, 1),
    (0, 2),
    (0, 3),
    (1, 3),
    (2, 1),
    (2, 2),
    (2, 3),
    (4, 1),
    }
    n_steps = decoder.validate(syndrome)
    assert n_steps == 26
    assert decoder.erasure == {
        ((0, 0), (0, 1)),
        ((0, 1), (0, 2)),
        ((0, 1), (1, 1)),
        ((0, 2), (0, 3)),
        ((0, 2), (1, 2)),
        ((0, 3), (0, 4)),
        ((0, 3), (1, 3)),
        ((1, 1), (2, 1)),
        ((1, 2), (1, 3)),
        ((1, 2), (2, 2)),
        ((1, 3), (1, 4)),
        ((1, 3), (2, 3)),
        ((2, 0), (2, 1)),
        ((2, 1), (2, 2)),
        ((2, 1), (3, 1)),
        ((2, 2), (2, 3)),
        ((2, 2), (3, 2)),
        ((2, 3), (2, 4)),
        ((2, 3), (3, 3)),
        ((3, 1), (4, 1)),
        ((4, 0), (4, 1)),
        ((4, 1), (4, 2)),
        ((4, 1), (5, 1))
    }

def test_validate_globe_shape(sf9F: Surface):
    decoder = Macar(sf9F)
    syndrome = {
        (0, 0),
        (0, 2),
        (0, 4),
        (0, 6),
        (2, 6),
        (3, 2),
        (4, 2),
        (4, 6),
        (5, 2),
        (6, 6),
        (8, 2),
        (8, 4),
        (8, 6),
    }
    n_steps = decoder.validate(syndrome)
    assert n_steps == 62
    assert decoder.erasure == {
        ((0, -1), (0, 0)),
        ((0, 0), (0, 1)),
        ((0, 0), (1, 0)),
        ((0, 1), (0, 2)),
        ((0, 2), (0, 3)),
        ((0, 2), (1, 2)),
        ((0, 3), (0, 4)),
        ((0, 4), (0, 5)),
        ((0, 4), (1, 4)),
        ((0, 5), (0, 6)),
        ((0, 6), (0, 7)),
        ((0, 6), (1, 6)),
        ((1, 2), (2, 2)),
        ((1, 6), (2, 6)),
        ((2, 1), (2, 2)),
        ((2, 1), (3, 1)),
        ((2, 2), (2, 3)),
        ((2, 2), (3, 2)),
        ((2, 3), (3, 3)),
        ((2, 5), (2, 6)),
        ((2, 6), (2, 7)),
        ((2, 6), (3, 6)),
        ((3, 0), (3, 1)),
        ((3, 1), (3, 2)),
        ((3, 1), (4, 1)),
        ((3, 2), (3, 3)),
        ((3, 2), (4, 2)),
        ((3, 3), (3, 4)),
        ((3, 3), (4, 3)),
        ((3, 6), (4, 6)),
        ((4, 0), (4, 1)),
        ((4, 1), (4, 2)),
        ((4, 1), (5, 1)),
        ((4, 2), (4, 3)),
        ((4, 2), (5, 2)),
        ((4, 3), (4, 4)),
        ((4, 3), (5, 3)),
        ((4, 5), (4, 6)),
        ((4, 6), (4, 7)),
        ((4, 6), (5, 6)),
        ((5, 0), (5, 1)),
        ((5, 1), (5, 2)),
        ((5, 1), (6, 1)),
        ((5, 2), (5, 3)),
        ((5, 2), (6, 2)),
        ((5, 3), (5, 4)),
        ((5, 3), (6, 3)),
        ((5, 6), (6, 6)),
        ((6, 1), (6, 2)),
        ((6, 2), (6, 3)),
        ((6, 2), (7, 2)),
        ((6, 5), (6, 6)),
        ((6, 6), (6, 7)),
        ((6, 6), (7, 6)),
        ((7, 2), (8, 2)),
        ((7, 4), (8, 4)),
        ((7, 6), (8, 6)),
        ((8, 1), (8, 2)),
        ((8, 2), (8, 3)),
        ((8, 3), (8, 4)),
        ((8, 4), (8, 5)),
        ((8, 5), (8, 6)),
        ((8, 6), (8, 7))
    }

def test_validate_single_defect(macar3T: Macar, actis3Tu: Actis):
    syndrome: set[Node] = {
        (0, 0, 0),
    }
    erasure = {
        ((0, -1, 0), (0, 0, 0)),
        ((0, 0, 0), (0, 0, 1)),
        ((0, 0, 0), (0, 1, 0)),
        ((0, 0, 0), (1, 0, 0))
    }

    n_steps = macar3T.validate(syndrome)
    assert n_steps == 10
    assert macar3T.erasure == erasure

    n_steps = actis3Tu.validate(syndrome)
    assert n_steps == 71
    assert actis3Tu.erasure == erasure

def test_validate_square(macar3T: Macar, actis3Tu: Actis):
    syndrome = {
        (0, 0, 0),
        (1, 0, 0),
        (0, 0, 1),
        (1, 0, 1),
    }
    erasure = {
        ((0, 0, 0), (0, 0, 1)),
        ((0, 0, 0), (1, 0, 0)),
        ((0, 0, 1), (1, 0, 1)),
        ((1, 0, 0), (1, 0, 1))
    }

    n_steps = macar3T.validate(syndrome)
    assert n_steps == 7
    assert macar3T.erasure == erasure

    n_steps = actis3Tu.validate(syndrome)
    assert n_steps == 35
    assert actis3Tu.erasure == erasure


def test_pointer_digraph(macar3F: Macar, uvw):
    u, v, _ = uvw
    macar3F.NODES.dc[u].pointer = 'W'
    macar3F.NODES.dc[v].pointer = 'W'
    macar3F.growth[u, v] = Growth.FULL
    dig, dig_diedges, dig_edges = macar3F._pointer_digraph
    assert type(dig) is nx.DiGraph
    assert set(dig.nodes) == set(macar3F.CODE.NODES)
    assert set(dig.edges) == {(v, u)}
    assert dig_diedges == [(v, u)]
    assert dig_edges == [(u, v)]


from localuf.constants import Growth
from localuf.decoders.uf import UF, _Cluster
from localuf.type_aliases import Node


class BUF(UF):
    """The graph used by Bucket UF.
    
    Extends ``decoders.UF``.
    
    Attributes:
    
    * ``mvl`` an integer of the minimum vision length of
        all active clusters at the start of a growth round.
    
    Methods:
    
    * ``_update_mvl``.
    * ``_demote_cluster``.
    
    Overriden attributes:
    
    * ``buckets`` replaces ``active_clusters`` and maps each integer
        to a set of all active clusters whose vision length is that integer.
        It is important that its keys is in ascending order
        as it is used by ``_update_mvl()``.
    
    Overriden methods:
    
    * ``reset``.
    * ``load``.
    * ``validate_syndrome``.
    * ``_merge``.
    * ``_update_self_after_union``.
    """

    def reset(self):
        super().reset()
        self.buckets: dict[int, set[_Cluster]] = {
            vision_length: set() for vision_length
            in range(1, self.CODE.N_EDGES+1)
        }
        self.mvl = None

    def load(self, syndrome):
        super().load(syndrome)
        for cluster in self.active_clusters:
            self.buckets[len(cluster.vision)].add(cluster)
        del self.active_clusters
        self._update_mvl()

    def _update_mvl(self):
        """Update ``mvl`` attribute."""
        changed = False
        for vision_length, bucket in self.buckets.items():
            if bucket:
                self.mvl = vision_length
                changed = True
                break
        if not changed:
            self.mvl = None

    # SYNDROME VALIDATION (need tests)

    def validate(
            self,
            syndrome: set[Node],
            log_history=False
    ):
        """Grow clusters (always shortest vision first) until they are all valid."""
        self.load(syndrome)
        if log_history: self.init_history()
        while self.mvl is not None:
            self._growth_round(
                log_history,
                clusters_to_grow=self.buckets[self.mvl]
            )
            self._update_mvl()

    def static_merge(self, cu, cv):
        if cu == cv:
            self._demote_cluster(cu)
        else:
            self._union(cu, cv)

    def dynamic_merge(self, cu, cv, e):
        if cu == cv:
            self.growth[e] = Growth.BURNT
            self._demote_cluster(cu)
        elif (cu.boundary and cv.boundary):
            self.growth[e] = Growth.BURNT
            self._demote_cluster(cu)
            self._demote_cluster(cv)
        else:
            self._union(cu, cv)

    def _demote_cluster(self, cluster: _Cluster):
        """If cluster active, move down a bucket as an edge has been removed from its vision. Else do nothing."""
        vision_length = len(cluster.vision)
        try:
            self.buckets[vision_length+1].remove(cluster)
            self.buckets[vision_length].add(cluster)
        except KeyError: pass

    def _update_self_after_union(self, larger, smaller, old_larger_vision_length):
        """Update attributes clusters and buckets after union of larger w/ smaller."""

        # DELETE SMALLER CLUSTER
        del self.clusters[smaller.root]
        # discard instead of remove as smaller may be a node not in syndrome
        self.buckets[len(smaller.vision)+1].discard(smaller)  # +1 as we already removed (u, v) from all visions

        # UPDATE buckets DICTIONARY
        # discard instead of remove as larger could have had boundary before union, hence was not active
        self.buckets[old_larger_vision_length].discard(larger)
        if larger.odd and not larger.boundary:
            self.buckets[len(larger.vision)].add(larger)


from unittest import mock

import pytest

from localuf.type_aliases import Node
from localuf._pairs import Pairs

FourNodes = tuple[Node, Node, Node, Node]


def test_reset(pairs: Pairs, uvwx: FourNodes):
    u, v, _, _ = uvwx
    pairs._dc = {u: v, v: u}
    pairs.reset()
    assert pairs._dc == {}


def test_contains(pairs: Pairs, uvwx: FourNodes):
    u, v, w, _ = uvwx
    pairs._dc = {u: v, v: u}
    assert u in pairs
    assert v in pairs
    assert w not in pairs


def test_getitem(pairs: Pairs, uvwx: FourNodes):
    u, v, w, _ = uvwx
    pairs._dc = {u: v, v: u}
    assert pairs[u] == v
    assert pairs[v] == u
    with pytest.raises(KeyError):
        pairs[w]


def test_add(pairs: Pairs, uvwx: FourNodes):
    u, v, _, _ = uvwx
    pairs.add(u, v)
    assert pairs._dc == {u: v, v: u}


def test_remove(pairs: Pairs, uvwx: FourNodes):
    u, v, _, _ = uvwx
    pairs._dc = {u: v, v: u}
    pairs.remove(u)
    assert pairs._dc == {}


def test_load_annihilates_anyons(pairs: Pairs, uvwx: FourNodes):
    u, v, _, _ = uvwx
    pairs._dc = {u: v, v: u}
    with mock.patch(
        "localuf._pairs.Pairs.remove",
        side_effect=pairs.remove,
    ) as mock_remove:
        pairs.load((u, v))
        mock_remove.assert_called_once_with(u)


def test_load_joins_paths(pairs: Pairs, uvwx: FourNodes):
    u, v, w, x = uvwx
    pairs._dc = {u: w, w: u, v: x, x: v}
    with (
        mock.patch("localuf._pairs.Pairs.remove") as mock_remove,
        mock.patch("localuf._pairs.Pairs.add") as mock_add,
    ):
        pairs.load((u, v))
        assert mock_remove.call_args_list == [mock.call(u), mock.call(v)]
        mock_add.assert_called_once_with(w, x)


def test_load_extends_path(pairs: Pairs, uvwx: FourNodes):
    u, v, w, _ = uvwx
    pairs._dc = {u: w, w: u}
    with (
        mock.patch("localuf._pairs.Pairs.remove") as mock_remove,
        mock.patch("localuf._pairs.Pairs.add") as mock_add,
    ):
        pairs.load((u, v))
        mock_remove.assert_called_once_with(u)
        mock_add.assert_called_once_with(v, w)
    # test symmetry
    with (
        mock.patch("localuf._pairs.Pairs.remove") as mock_remove,
        mock.patch("localuf._pairs.Pairs.add") as mock_add,
    ):
        pairs.load((v, u))
        mock_remove.assert_called_once_with(u)
        mock_add.assert_called_once_with(v, w)


def test_load_adds_path(pairs: Pairs, uvwx: FourNodes):
    u, v, _, _ = uvwx
    with mock.patch("localuf._pairs.Pairs.add") as mock_add:
        pairs.load((u, v))
        mock_add.assert_called_once_with(u, v)


def test_as_set(pairs: Pairs, uvwx: FourNodes):
    u, v, w, x = uvwx
    pairs._dc = {u: v, v: u, w: x, x: w}
    assert pairs.as_set == {(u, v), (w, x)}


import pytest

from localuf.decoders.luf.constants import Stage

def test_str():
    assert str(Stage.GROWING) == 'G'
    assert str(Stage.MERGING) == 'M'
    assert str(Stage.PRESYNCING) == 'PS'
    assert str(Stage.SYNCING) == 'S'
    assert str(Stage.BURNING) == 'B'
    assert str(Stage.PEELING) == 'P'
    assert str(Stage.DONE) == 'D'

def test_iadd():
    x = Stage.GROWING
    x += Stage.INCREMENT
    assert x is Stage.MERGING
    x += Stage.INCREMENT
    assert x is Stage.PRESYNCING
    x += Stage.INCREMENT
    assert x is Stage.SYNCING
    x += Stage.INCREMENT
    assert x is Stage.BURNING
    x += Stage.INCREMENT
    assert x is Stage.PEELING
    x += Stage.INCREMENT
    assert x is Stage.DONE
    with pytest.raises(ValueError, match="7 is not a valid Stage"):
        x += Stage.INCREMENT

def test_imod():
    x = Stage.BURNING
    x %= Stage.SV_STAGE_COUNT
    assert x is Stage.GROWING